from collections.abc import Sequence

import numpy as np
import pandas as pd

__all__ = ["match_heartbeat_lists"]

from scipy.spatial import KDTree, minkowski_distance


def match_heartbeat_lists(
    *,
    heartbeats_reference: pd.DataFrame,
    heartbeats_extracted: pd.DataFrame,
    tolerance_ms: int | float = 10,
    sampling_rate_hz: float,
    one_to_one: bool = True,
    heartbeats_reference_suffix: str | None = "_reference",
    heartbeats_extracted_suffix: str | None = "",
) -> pd.DataFrame:
    """Find True Positives, False Positives and True Negatives by comparing extracted heartbeats with ground truth.

    This compares a extracted heartbeat list with a ground truth heartbeat list and returns True Positives,
    False Positives and True Negatives matches.
    The comparison is purely based on the start and end values of each heartbeat in the lists.
    Two heartbeat are considered a positive match, if both their start and their end values differ by less than the
    `threshold`.

    By default (controlled by the one-to-one parameter), if multiple heartbeats of the extracted heartbeat list would
    match to a single ground truth heartbeat (or vise-versa), only the heartbeat with the lowest distance is considered
    an actual match.
    If `one_to_one` is set to False, all matches would be considered True positives.
    This might lead to unexpected results in certain cases and should not be used to calculate traditional metrics like
    precision and recall.

    It is highly recommended to order the heartbeat lists and remove heartbeat with large overlaps before applying this
    method to get reliable results.

    Parameters
    ----------
    heartbeats_reference : :class:`~pandas.DataFrame`
        The ground truth heartbeat list.
    heartbeats_extracted : :class:`~pandas.DataFrame`
        The list of extracted heartbeats.
    sampling_rate_hz : int or float
        The sampling rate of the ECG signal in Hz.
    tolerance_ms : int or float, optional
        The allowed tolerance between labels in milliseconds.
        The comparison is done as `distance <= tolerance_ms`.
    one_to_one : bool, optional
        If True, only a single unique match per heartbeat is considered.
        In case of multiple matches, the one with the lowest distance is considered.
        If case of multiple matches with the same distance, the first match will be considered.
        If False, multiple matches are possible.
        If this is set to False, some calculated metrics from these matches might not be well defined!
    heartbeats_extracted_suffix : str, optional
        A suffix that will be appended to the index name of the extracted stride list in the output.
    heartbeats_reference_suffix : str, optional
        A suffix that will be appended to the index name of the ground truth in the output.

    Returns
    -------
    matches
        A 3 column dataframe with the column names `heartbeat_id{heartbeats_extracted_suffix}`,
        `heartbeat_id{heartbeats_reference_suffix}`, and `match_type`.
        Each row is a match containing the index value of the left and the right list, that belong together.
        The `match_type` column indicates the type of match.
        For all extracted heartbeats that have a match in the ground truth list, this will be "tp" (true positive).
        Extracted heartbeats that do not have a match will be mapped to a NaN and the match-type will be "fp" (false
        positives)
        All ground truth strides that do not have a extracted counterpart are marked as "fn" (false negative).
        In case MultiSensorStrideLists were used as inputs, a dictionary of such dataframes is returned.


    Examples
    --------
    >>> heartbeat_reference = pd.DataFrame(
    ...     [[10, 21], [20, 34], [31, 40]], columns=["start_sample", "end_sample"]
    ... ).rename_axis("heartbeat_id")
    >>> stride_list_seg = pd.DataFrame(
    ...     [[10, 20], [21, 30], [31, 40], [50, 60]], columns=["start_sample", "end_sample"]
    ... ).rename_axis("heartbeat_id")
    >>> matches = match_heartbeat_lists(
    ...     heartbeats_reference=heartbeats_reference,
    ...     heartbeats_extracted=heartbeats_extracted,
    ...     sampling_rate_hz=500,
    ...     tolerance_ms=10,
    ... )
    >>> matches
      heartbeat_id heartbeat_id_reference match_type
    0    0                 0         tp
    1    1               NaN         fp
    2    2                 2         tp
    3    3               NaN         fp
    4  NaN                 1         fn

    """
    tolerance_samples = int(tolerance_ms / 1000 * sampling_rate_hz)
    matches = _match_heartbeat_lists(
        heartbeats_extracted,
        heartbeats_reference,
        match_cols=["start_sample", "end_sample"],
        tolerance_samples=tolerance_samples,
        one_to_one=one_to_one,
        suffix_a=heartbeats_extracted_suffix,
        suffix_b=heartbeats_reference_suffix,
    )

    segmented_index_name = heartbeats_extracted.index.name + heartbeats_extracted_suffix
    reference_index_name = heartbeats_reference.index.name + heartbeats_reference_suffix

    tp_idx = ~matches.isna().any(axis=1)
    matches = matches.assign(match_type=pd.NA)
    matches.loc[tp_idx, "match_type"] = "tp"
    matches.loc[matches[reference_index_name].isna(), "match_type"] = "fp"
    matches.loc[matches[segmented_index_name].isna(), "match_type"] = "fn"

    return matches


def _match_heartbeat_lists(
    heartbeat_list_a: pd.DataFrame,
    heartbeat_list_b: pd.DataFrame,
    match_cols: str | Sequence[str],
    tolerance_samples: int | float = 0,
    one_to_one: bool = True,
    suffix_a: str = "_a",
    suffix_b: str = "_b",
) -> pd.DataFrame:
    if suffix_a == suffix_b:
        raise ValueError("The suffix for the first and the second heartbeat list must be different.")

    if tolerance_samples < 0:
        raise ValueError("The tolerance must be larger 0.")

    # sanitize input
    match_cols = [match_cols] if isinstance(match_cols, str) else list(match_cols)

    left_indices, right_indices = _match_label_lists(
        heartbeat_list_a[match_cols].to_numpy(),
        heartbeat_list_b[match_cols].to_numpy(),
        tolerance_samples=tolerance_samples,
        one_to_one=one_to_one,
    )

    index_name_a = heartbeat_list_a.index.name + suffix_a
    index_name_b = heartbeat_list_b.index.name + suffix_b

    matches_a = pd.DataFrame(index=heartbeat_list_a.index.copy(), columns=[index_name_b])
    matches_a.index.name = index_name_a

    matches_b = pd.DataFrame(index=heartbeat_list_b.index.copy(), columns=[index_name_a])
    matches_b.index.name = index_name_b

    heartbeat_list_a_idx = heartbeat_list_a.iloc[left_indices].index
    heartbeat_list_b_idx = heartbeat_list_b.iloc[right_indices].index

    matches_a.loc[heartbeat_list_a_idx, index_name_b] = heartbeat_list_b_idx
    matches_b.loc[heartbeat_list_b_idx, index_name_a] = heartbeat_list_a_idx

    matches_a = matches_a.reset_index()
    matches_b = matches_b.reset_index()

    matches = (
        pd.concat([matches_a, matches_b])
        .drop_duplicates()
        .sort_values([index_name_a, index_name_b])
        .reset_index(drop=True)
    )

    return matches


def _match_label_lists(
    list_a: np.ndarray, list_b: np.ndarray, tolerance_samples: int, one_to_one: bool
) -> tuple[np.ndarray, np.ndarray]:
    """Find matches in two lists based on the distance between their vectors.

    Parameters
    ----------
    list_a : array with shape (n, d)
        An n long array of d-dimensional vectors
    list_b : array with shape (m, d)
        An m long array of d-dimensional vectors
    tolerance_samples
        Max allowed Chebyshev distance between matches.
        The comparison is done as "distance <= tolerance".
    one_to_one
        If True only valid one-to-one matches are returned (see more below)

    Returns
    -------
    indices_a
        Indices from list a that have a match in the list b.
        If `one_to_one` is False, indices might repeat.
    indices_b
        Indices from list b that have a match in list b.
        If `one_to_one` is False, indices might repeat.
        A valid match pare is then `(indices_b[i], indices_b[i]) for all i.

    Notes
    -----
    This function supports 2 modes:

    `one_to_one` = False:
        In this mode every match is returned as long the distance in all dimensions between the matches is at most
        tolerance.
        This is equivalent to the Chebyshev distance between the matches
        (aka `np.max(np.abs(match_a - match_b)) < tolerance`).
        This means multiple matches for each vector will be returned.
        This means the respective indices will occur multiple times in the output vectors.
    `one_to_one` = True:
        In this mode only a single match per index is allowed in both directions.
        This means that every index will only occur once in the output arrays.
        If multiple matches are possible based on the tolerance of the Chebyshev distance, the closest match will be
        selected based on the Manhattan distance (aka `np.sum(np.abs(match_a - match_b`).
        Only this match will be returned.
        This is done, because in case the input arrays are multi-dimensional, the Chebyshev distance is not really
        well suited for comparison.

    """
    if len(list_a) == 0 or len(list_b) == 0:
        return np.array([]), np.array([])

    tree_a = KDTree(list_b)
    tree_b = KDTree(list_a)

    if one_to_one is False:
        # p = np.inf is used to select the Chebyshev distance
        keys = list(
            zip(
                # We force sort the keys here to make sure the order is deterministic, even if the storage order of
                # sparse array might not be.
                *sorted(
                    tree_a.sparse_distance_matrix(tree_b, tolerance_samples, p=np.inf, output_type="dict").keys(),
                    key=lambda x: x[1],
                ),
                strict=False,
            )
        )
        # All values are returned that have a valid match
        return (np.array([]), np.array([])) if len(keys) == 0 else (np.array(keys[1]), np.array(keys[0]))

    # one_to_one is True
    # We calculate the closest neighbor based on the Manhattan distance in both directions and then find only the cases
    # were the right side closest neighbor resulted in the same pairing as side a closest neighbor ensuring
    # that we have true one-to-one-matches and we have already the closest match based on our final criteria.

    # p = 1 is used to select the Manhattan distance
    nearest_distance_a, nearest_neighbor_a = tree_a.query(list_a, p=1, workers=-1)
    _, nearest_neighbor_b = tree_b.query(list_b, p=1, workers=-1)

    # Filter the once that are true one-to-one matches
    indices_a = np.arange(len(list_a))
    combined_indices = np.vstack([indices_a, nearest_neighbor_a]).T
    boolean_map = nearest_neighbor_b[nearest_neighbor_a] == indices_a
    valid_matches = combined_indices[boolean_map]

    # Check if the remaining matches are inside our Chebyshev tolerance distance.
    # If not, delete them.
    valid_matches_distance = nearest_distance_a[boolean_map]
    # First we check if any of the Manhattan distances is larger than the threshold.
    # If not, all the Chebyshev distances are smaller than the threshold, too.
    index_large_matches = np.where(~(valid_matches_distance <= tolerance_samples))[0]
    if index_large_matches.size > 0:
        # Minkowski with p = np.inf uses the Chebyshev distance
        boolean_map = (
            minkowski_distance(list_a[index_large_matches], list_b[valid_matches[index_large_matches, 1]], p=np.inf)
            <= tolerance_samples
        )

        valid_matches = np.delete(valid_matches, index_large_matches[~boolean_map], axis=0)

    valid_matches = valid_matches.T

    return valid_matches[0], valid_matches[1]



import numpy as np

from pepbench.utils._types import str_t

__all__ = [
    "get_nan_reason_mapping",
    "rename_algorithms",
    "rename_metrics",
]

_ylabel_mapping = {
    "pep_ms": "PEP [ms]",
    "rr_interval_ms": "RR-Interval [ms]",
    "error_per_sample_ms": "Error [ms]",
    "absolute_error_per_sample_ms": "Absolute Error [ms]",
    "absolute_relative_error_per_sample_percent": "Absolute Relative Error [%]",
}

_xlabel_mapping = {
    "phase": "Phase",
    "participant": "Participant",
    "condition": "Condition",
}

_algo_level_mapping = {
    "q_peak_algorithm": "Q-Peak Detection",
    "b_point_algorithm": "B-Point Detection",
    "outlier_correction_algorithm": "Outlier Correction",
}

_algorithm_mapping = {
    "b-point-reference": "Ref. B-Point",
    "q-peak-reference": "Ref. Q-Peak",
    "none": "None",
    "stern1985": "Ste85",
    "sherwood1990": "She90",
    "debski1993-second-derivative": "Deb93SD",
    "martinez2004": "Mar04",
    "lozano2007-linear-regression": "Loz07LR",
    "lozano2007-quadratic-regression": "Loz07QR",
    "arbol2017-isoelectric-crossings": "Arb17IC",
    "arbol2017-second-derivative": "Arb17SD",
    "arbol2017-third-derivative": "Arb17TD",
    "forouzanfar2018": "For18",
    "miljkovic2022": "Mil22",
    "pale2021": "Pal21",
    "drost2022": "Dro22",
    "linear-interpolation": "LinInt",
}
_algorithm_mapping.update(**{f"vanlien2013-{i}-ms": f"Van13 ({i} ms)" for i in np.arange(32, 44, 2)})
_algorithm_mapping.update()

_metric_mapping = {
    "mean": "Mean",
    "std": "SD",
    "total": "Total",
    "absolute_error_per_sample_ms": "Absolute Error [ms]",
    "error_per_sample_ms": "Error [ms]",
    "relative_error_per_sample_percent": "Relative Error [%]",
}

_nan_reason_mapping = {
    "c_point_nan": "No C-point detected",
    "r_peak_nan": "No R-peak detected",
    "negative_pep": "Negative PEP",
    "no_iso_crossing_before_c_point": "No Isoelectric Crossing before C-point detected",
    "no_local_minimum": "No Local Minimum detected",
    "no_monotonic_segment": "No Monotonically Increasing Segment detected",
    "no_zero_crossing": "No Zero Crossings detected",
    "invalid_b_point_search_window": "Invalid B-point Search Window",
}

_nan_reason_mapping_short = {
    "c_point_nan": "No C",
    "r_peak_nan": "No R",
    "negative_pep": "Neg. PEP",
    "no_iso_crossing_before_c_point": "No IC",
    "no_local_minimum": "No Loc. Min.",
    "no_monotonic_segment": "No Mon. Incr.",
    "no_zero_crossing": "No ZCs",
    "invalid_b_point_search_window": "Inv. B Window",
}


def rename_metrics(metrics: str_t) -> str_t:
    if isinstance(metrics, str):
        return _metric_mapping.get(metrics, metrics)
    return [_metric_mapping.get(metric, metric) for metric in metrics]


def rename_algorithms(algorithms: str_t) -> str_t:
    if isinstance(algorithms, str):
        return _algorithm_mapping.get(algorithms, algorithms)
    return [_algorithm_mapping.get(algo, algo) for algo in algorithms]


def get_nan_reason_mapping() -> dict:
    return {_nan_reason_mapping_short[k]: _nan_reason_mapping[k] for k in _nan_reason_mapping}



import contextlib
import time
from collections.abc import Generator
from datetime import datetime
from typing import TypedDict


class MeasureTimeResults(TypedDict):
    """Results of the measure_time context manager."""

    start_datetime_utc_timestamp: float
    start_datetime: str
    end_start_datetime_utc_timestamp: float
    end_start_datetime: str
    runtime: float


@contextlib.contextmanager
def measure_time() -> Generator[MeasureTimeResults, None, None]:
    """Context manager to measure the execution time.

    Note: This is not meant for high precision timing.
    """
    results = {
        "start_datetime_utc_timestamp": datetime.utcnow().timestamp(),
        "start_datetime": datetime.now().astimezone().isoformat(),
    }
    start_time = time.perf_counter()
    yield results
    end_time = time.perf_counter()
    results["end_datetime_utc_timestamp"] = datetime.utcnow().timestamp()
    results["end_datetime"] = datetime.now().astimezone().isoformat()
    results["runtime_s"] = end_time - start_time



from typing import TYPE_CHECKING, Literal, TypeVar

import pandas as pd
from biopsykit.signals._base_extraction import HANDLE_MISSING_EVENTS
from biopsykit.signals.ecg.event_extraction import BaseEcgExtraction
from biopsykit.signals.ecg.segmentation import BaseHeartbeatSegmentation
from biopsykit.signals.icg.event_extraction import (
    BaseBPointExtraction,
    BaseCPointExtraction,
    CPointExtractionScipyFindPeaks,
)
from biopsykit.signals.icg.outlier_correction import BaseBPointOutlierCorrection, OutlierCorrectionDummy
from biopsykit.signals.pep import PepExtraction
from biopsykit.signals.pep._pep_extraction import NEGATIVE_PEP_HANDLING
from biopsykit.utils.dtypes import (
    BPointDataFrame,
    CPointDataFrame,
    HeartbeatSegmentationDataFrame,
    PepResultDataFrame,
    QPeakDataFrame,
    is_pep_result_dataframe,
)
from tpcp import CloneFactory, Parameter, Pipeline

from pepbench._docutils import make_filldoc

if TYPE_CHECKING:
    from pepbench.datasets import BasePepDataset, BasePepDatasetWithAnnotations

__all__ = ["BasePepExtractionPipeline"]


BasePepDatasetT = TypeVar("BasePepDatasetT", bound="BasePepDataset")
BasePepDatasetWithAnnotationsT = TypeVar("BasePepDatasetWithAnnotationsT", bound="BasePepDatasetWithAnnotations")

base_pep_pipeline_docfiller = make_filldoc(
    {
        "base_parameters": """
        Parameters
        ----------
        heartbeat_segmentation_algo : :class:`~biopsykit.signals.ecg.segmentation.BaseHeartbeatSegmentation`
            Algorithm for heartbeat segmentation.
        q_peak_algo : :class:`~biopsykit.signals.ecg.event_extraction.BaseEcgExtraction`
            Algorithm for Q-peak extraction.
        b_point_algo : :class:`~biopsykit.signals.icg.event_extraction.BaseBPointExtraction`
            Algorithm for B-point extraction.
        c_point_algo : :class:`~biopsykit.signals.icg.event_extraction.BaseCPointExtraction`
            Algorithm for C-point extraction, necessary for most subsequent B-point extraction algorithms.
        outlier_correction_algo : :class:`~biopsykit.signals.icg.outlier_correction.BaseOutlierCorrection`
            Algorithm for outlier correction of B-point data (optional).
        handle_negative_pep : one of {`"nan"`, `"zero"`, `"keep"`}
            How to handle negative PEP values. Possible values are:
                - `"nan"`: Set negative PEP values to NaN
                - `"zero"`: Set negative PEP values to 0
                - `"keep"`: Keep negative PEP values as is
        handle_missing_events : one of {`"warn"`, `"ignore"`, `"raise"`}
            How to handle missing events. Possible values are:
                - `"warn"`: Issue a warning if missing events are detected
                - `"ignore"`: Ignore missing events
                - `"raise"`: Raise an error if missing events are detected
        """,
        "datapoint_pipeline": """
        datapoint : :class:`~pepbench.datasets._base_pep_extraction_dataset.BasePepDataset`
            The data to run the pipeline on. This needs to be a valid datapoint (i.e. a dataset with just a single row).
            The Dataset should be a child class of
            :class:`~pepbench.datasets._base_pep_extraction_dataset.BasePepDataset` or implement all the same
            parameters and methods.
        """,
        "datapoint_pipeline_labeled": """
        datapoint : :class:`~pepbench.datasets._base_pep_extraction_dataset.BaseUnifiedPepExtractionDataset`
            The data to run the pipeline on. This needs to be a valid datapoint (i.e. a dataset with just a single row).
            The Dataset should be a child class of
            :class:`~pepbench.datasets._base_pep_extraction_dataset.BaseUnifiedPepExtractionDataset` or implement all
            the same parameters and methods. This means that it must *also* implement methods to get the reference
            heartbeats and reference PEP.
            """,
        "attributes": """
        Attributes
        ----------
        heartbeat_segmentation_results_ : :class:`~biopsykit.signals.ecg.segmentation.HeartbeatSegmentationDataFrame`
            Results from the heartbeat segmentation step.
        q_peak_results_ : :class:`~biopsykit.signals.ecg.event_extraction.QPeakDataFrame`
            Results from the Q-peak extraction step.
        c_point_results_ : :class:`~biopsykit.signals.icg.event_extraction.CPointDataFrame`
            Results from the C-point extraction step.
        b_point_results_ : :class:`~biopsykit.signals.icg.event_extraction.BPointDataFrame`
            Results from the B-point extraction step.
        b_point_after_outlier_correction_results_ : :class:`~biopsykit.signals.icg.event_extraction.BPointDataFrame`
            Results from the B-point extraction step after outlier correction.
        pep_results_ : :class:`~biopsykit.signals.pep.PepResultDataFrame`
            Results from the PEP extraction step.
        """,
    },
    doc_summary="Decorator to fill common parts of the docstring for subclasses of :class:`BasePepExtractionPipeline`.",
)


@base_pep_pipeline_docfiller
class BasePepExtractionPipeline(Pipeline):
    """Base class for PEP extraction pipelines.

    This class provides all the necessary methods to extract PEP from ECG and ICG data using the specified algorithms.
    For usage, it is recommended to use the derived pipelines
    (e.g., :class:`~pepbench.pipelines.PepExtractionPipeline`).

    %(base_parameters)s

    %(attributes)s

    """

    heartbeat_segmentation_algo: Parameter[BaseHeartbeatSegmentation]
    q_peak_algo: Parameter[BaseEcgExtraction]
    b_point_algo: Parameter[BaseBPointExtraction]
    c_point_algo: Parameter[BaseCPointExtraction]
    outlier_correction_algo: Parameter[BaseBPointOutlierCorrection]
    handle_negative_pep: NEGATIVE_PEP_HANDLING
    handle_missing_events: HANDLE_MISSING_EVENTS

    heartbeat_segmentation_results_: HeartbeatSegmentationDataFrame
    q_peak_results_: QPeakDataFrame
    c_point_results_: CPointDataFrame | None
    b_point_results_: BPointDataFrame
    b_point_after_outlier_correction_results_: BPointDataFrame
    pep_results_: PepResultDataFrame

    def __init__(
        self,
        *,
        heartbeat_segmentation_algo: BaseHeartbeatSegmentation,
        q_peak_algo: BaseEcgExtraction,
        b_point_algo: BaseBPointExtraction,
        c_point_algo: BaseCPointExtraction = CloneFactory(CPointExtractionScipyFindPeaks()),
        outlier_correction_algo: BaseBPointOutlierCorrection | None = None,
        handle_negative_pep: Literal[NEGATIVE_PEP_HANDLING] = "nan",
        handle_missing_events: Literal[HANDLE_MISSING_EVENTS] | None = None,
    ) -> None:
        self.heartbeat_segmentation_algo = heartbeat_segmentation_algo
        self.q_peak_algo = q_peak_algo
        self.b_point_algo = b_point_algo
        self.c_point_algo = c_point_algo
        if outlier_correction_algo is None:
            outlier_correction_algo = OutlierCorrectionDummy()
        self.outlier_correction_algo = outlier_correction_algo
        self.pep_extraction_algo = PepExtraction()
        self.handle_negative_pep = handle_negative_pep

        if handle_missing_events is None:
            handle_missing_events = "warn"
        self.handle_missing_events = handle_missing_events

    def _compute_pep(
        self,
        *,
        heartbeats: HeartbeatSegmentationDataFrame,
        q_peak_samples: QPeakDataFrame,
        b_point_samples: BPointDataFrame,
        sampling_rate_hz: float,
    ) -> pd.DataFrame:
        pep_extraction_algo = PepExtraction(handle_negative_pep=self.handle_negative_pep)
        pep_extraction_algo.extract(
            heartbeats=heartbeats,
            q_peak_samples=q_peak_samples,
            b_point_samples=b_point_samples,
            sampling_rate_hz=sampling_rate_hz,
        )

        pep_results = pep_extraction_algo.pep_results_.copy()
        pep_results = pep_results.astype(
            {
                "heartbeat_start_sample": "Int64",
                "heartbeat_end_sample": "Int64",
                "r_peak_sample": "Int64",
                "rr_interval_sample": "Int64",
                "rr_interval_ms": "Float64",
                "heart_rate_bpm": "Float64",
                "q_peak_sample": "Int64",
                "b_point_sample": "Int64",
                "pep_sample": "Int64",
                "pep_ms": "Float64",
                "nan_reason": "object",
            }
        )

        is_pep_result_dataframe(pep_results)

        return pep_results



from typing import get_args

import pandas as pd
from biopsykit.signals._base_extraction import CanHandleMissingEventsMixin
from biopsykit.signals.pep._pep_extraction import NEGATIVE_PEP_HANDLING
from typing_extensions import Self

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.heartbeat_matching import match_heartbeat_lists
from pepbench.pipelines._base_pipeline import BasePepExtractionPipeline, base_pep_pipeline_docfiller

__all__ = ["PepExtractionPipelineReferenceBPoints"]


@base_pep_pipeline_docfiller
class PepExtractionPipelineReferenceBPoints(BasePepExtractionPipeline):
    """`tpcp` Pipeline for PEP extraction that uses reference B-points for B-point detection.

    This pipeline is used to validate different Q-peak extraction algorithms and computing the PEP using reference
    B-points.

    %(base_parameters)s

    Other Parameters
    ----------------
    %(datapoint_pipeline_labeled)s

    %(attributes)s

    """

    @base_pep_pipeline_docfiller
    def run(self, datapoint: BasePepDatasetWithAnnotations) -> Self:
        """Run the pipeline on the given datapoint.

        The pipeline will extract PEP from the given datapoint using the specified algorithms. The results will be
        stored in the attributes of the class.

        Parameters
        ----------
        %(datapoint_pipeline_labeled)s

        """
        if self.handle_negative_pep not in get_args(NEGATIVE_PEP_HANDLING):
            raise ValueError(
                f"Invalid value for 'handle_negative_pep': {self.handle_negative_pep}. "
                f"Must be one of {NEGATIVE_PEP_HANDLING}"
            )

        heartbeat_algo = self.heartbeat_segmentation_algo.clone()
        q_peak_algo = self.q_peak_algo.clone()
        outlier_algo = self.outlier_correction_algo.clone()

        reference_pep = datapoint.reference_pep
        fs_ecg = datapoint.sampling_rate_ecg

        ecg_data = datapoint.ecg

        # set handle_missing parameter for all algorithms
        if self.handle_missing_events is not None:
            for algo in (heartbeat_algo, q_peak_algo, outlier_algo):
                if isinstance(algo, CanHandleMissingEventsMixin):
                    # this overwrites the default value of the handle_missing parameter
                    algo.set_params(handle_missing_events=self.handle_missing_events)

        # extract heartbeats
        heartbeat_algo.extract(ecg=ecg_data, sampling_rate_hz=fs_ecg)
        heartbeats = heartbeat_algo.heartbeat_list_

        # run Q-peak extraction
        q_peak_algo.extract(ecg=ecg_data, heartbeats=heartbeats, sampling_rate_hz=fs_ecg)
        q_peak_samples = q_peak_algo.points_

        heartbeat_matching = match_heartbeat_lists(
            heartbeats_reference=datapoint.reference_heartbeats,
            heartbeats_extracted=heartbeats,
            tolerance_ms=100,
            sampling_rate_hz=datapoint.sampling_rate_ecg,
        )

        # TODO: handle false negatives and false positives, i.e. heartbeats that are not matched
        tp_matches = heartbeat_matching.query("match_type == 'tp'")
        q_peak_samples_tp = q_peak_samples.loc[tp_matches["heartbeat_id"]]
        b_point_samples = reference_pep[["b_point_sample", "nan_reason"]].copy()
        b_point_samples_tp = b_point_samples.loc[tp_matches["heartbeat_id_reference"]]

        tp_matches = tp_matches.set_index("heartbeat_id")

        q_peak_samples_tp.index = tp_matches.index
        b_point_samples_tp.index = tp_matches.index

        # add nan_reason column to match extracted b-points
        b_point_samples_tp["nan_reason"] = pd.NA

        outlier_algo.correct_outlier(
            b_points=b_point_samples_tp,
            c_points=None,
            sampling_rate_hz=0,
            handle_missing=self.handle_missing_events,
            handle_negative=self.handle_negative_pep,
        )
        b_point_samples_after_outlier = outlier_algo.points_

        pep_results = self._compute_pep(
            heartbeats=heartbeats,
            q_peak_samples=q_peak_samples_tp,
            b_point_samples=b_point_samples_after_outlier,
            sampling_rate_hz=fs_ecg,
        )

        self.heartbeat_segmentation_results_ = heartbeats
        self.q_peak_results_ = q_peak_samples
        self.c_point_results_ = None
        self.b_point_results_ = b_point_samples
        self.b_point_after_outlier_correction_results_ = b_point_samples_after_outlier
        self.pep_results_ = pep_results
        return self



from typing import get_args

import pandas as pd
from biopsykit.signals._base_extraction import CanHandleMissingEventsMixin
from biopsykit.signals.pep._pep_extraction import NEGATIVE_PEP_HANDLING
from tpcp._dataset import DatasetT
from typing_extensions import Self

from pepbench.heartbeat_matching import match_heartbeat_lists
from pepbench.pipelines._base_pipeline import BasePepExtractionPipeline, base_pep_pipeline_docfiller

__all__ = ["PepExtractionPipelineReferenceQPeaks"]


@base_pep_pipeline_docfiller
class PepExtractionPipelineReferenceQPeaks(BasePepExtractionPipeline):
    """`tpcp` Pipeline for PEP extraction that uses reference Q-peaks for Q-peak detection.

    This pipeline is used to validate different B-point extraction algorithms and computing the PEP using reference
    Q-peaks.

    %(base_parameters)s

    Other Parameters
    ----------------
    %(datapoint_pipeline_labeled)s

    %(attributes)s

    """

    def run(self, datapoint: DatasetT) -> Self:
        """Run the pipeline on the given datapoint.

        The pipeline will extract PEP from the given datapoint using the specified algorithms. The results will be
        stored in the attributes of the class.

        Parameters
        ----------
        %(datapoint_pipeline_labeled)s

        """
        if self.handle_negative_pep not in get_args(NEGATIVE_PEP_HANDLING):
            raise ValueError(
                f"Invalid value for 'handle_negative_pep': {self.handle_negative_pep}. "
                f"Must be one of {NEGATIVE_PEP_HANDLING}"
            )

        heartbeat_algo = self.heartbeat_segmentation_algo.clone()
        c_point_algo = self.c_point_algo.clone()
        b_point_algo = self.b_point_algo.clone()
        outlier_algo = self.outlier_correction_algo.clone()

        reference_pep = datapoint.reference_pep
        fs_ecg = datapoint.sampling_rate_ecg
        fs_icg = datapoint.sampling_rate_icg

        ecg_data = datapoint.ecg
        icg_data = datapoint.icg

        # set handle_missing parameter for all algorithms
        if self.handle_missing_events is not None:
            for algo in (heartbeat_algo, c_point_algo, b_point_algo, outlier_algo):
                if isinstance(algo, CanHandleMissingEventsMixin):
                    # this overwrites the default value of the handle_missing parameter
                    algo.set_params(handle_missing_events=self.handle_missing_events)

        # extract heartbeats
        heartbeat_algo.extract(ecg=ecg_data, sampling_rate_hz=fs_ecg)
        heartbeats = heartbeat_algo.heartbeat_list_

        heartbeat_matching = match_heartbeat_lists(
            heartbeats_reference=datapoint.reference_heartbeats,
            heartbeats_extracted=heartbeats,
            tolerance_ms=100,
            sampling_rate_hz=datapoint.sampling_rate_ecg,
        )

        # run C-point extraction
        c_point_algo.extract(icg=icg_data, heartbeats=heartbeats, sampling_rate_hz=fs_icg)

        # run B-point extraction
        b_point_algo.extract(
            icg=icg_data, heartbeats=heartbeats, c_points=c_point_algo.points_, sampling_rate_hz=fs_icg
        )

        # TODO: handle false negatives and false positives, i.e. heartbeats that are not matched
        tp_matches = heartbeat_matching.query("match_type == 'tp'")

        # run Q-Peak extraction
        q_peak_samples = reference_pep[["q_peak_sample"]].copy()
        q_peak_samples_tp = q_peak_samples.loc[tp_matches["heartbeat_id_reference"]]

        b_point_samples = b_point_algo.points_
        b_point_samples_tp = b_point_samples.loc[tp_matches["heartbeat_id"]]
        c_point_samples = c_point_algo.points_
        c_point_samples_tp = c_point_samples.loc[tp_matches["heartbeat_id"]]

        tp_matches = tp_matches.set_index("heartbeat_id")

        q_peak_samples_tp.index = tp_matches.index
        c_point_samples_tp.index = tp_matches.index
        b_point_samples_tp.index = tp_matches.index

        # add nan_reason column to match extracted b-points
        q_peak_samples_tp["nan_reason"] = pd.NA

        outlier_algo.correct_outlier(b_points=b_point_samples_tp, c_points=c_point_samples_tp, sampling_rate_hz=fs_icg)
        b_point_samples_after_outlier = outlier_algo.points_

        pep_results = self._compute_pep(
            heartbeats=heartbeats,
            q_peak_samples=q_peak_samples_tp,
            b_point_samples=b_point_samples_after_outlier,
            sampling_rate_hz=fs_icg,
        )

        self.heartbeat_segmentation_results_ = heartbeats
        self.q_peak_results_ = q_peak_samples_tp
        self.c_point_results_ = c_point_samples_tp
        self.b_point_results_ = b_point_samples_tp
        self.b_point_after_outlier_correction_results_ = b_point_samples_after_outlier
        self.pep_results_ = pep_results
        return self



"""Preprocess ECG and ICG signals."""

from biopsykit.signals.ecg.preprocessing import EcgPreprocessingNeurokit
from biopsykit.signals.icg.preprocessing import IcgPreprocessingBandpass

__all__ = ["EcgPreprocessingNeurokit", "IcgPreprocessingBandpass"]



from collections.abc import Sequence
from functools import cached_property, lru_cache
from itertools import product
from pathlib import Path
from typing import ClassVar

import pandas as pd
from biopsykit.metadata import bmi
from biopsykit.signals.ecg.preprocessing import EcgPreprocessingNeurokit
from biopsykit.signals.ecg.segmentation import HeartbeatSegmentationNeurokit
from biopsykit.signals.icg.preprocessing import IcgPreprocessingBandpass
from biopsykit.utils.dtypes import EcgRawDataFrame, HeartbeatSegmentationDataFrame, IcgRawDataFrame
from biopsykit.utils.file_handling import get_subject_dirs

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.datasets._base_pep_extraction_dataset import MetadataMixin, base_pep_extraction_docfiller
from pepbench.datasets._helper import compute_reference_heartbeats, load_labeling_borders
from pepbench.datasets.empkins._helper import _load_biopac_data, _load_timelog
from pepbench.utils._types import path_t

_cached_get_biopac_data = lru_cache(maxsize=4)(_load_biopac_data)
# cache_dir = "./cachedir"
# memory = Memory(location=cache_dir, verbose=0)
# _cached_get_biopac_data = memory.cache(_load_biopac_data)


@base_pep_extraction_docfiller
class EmpkinsDataset(BasePepDatasetWithAnnotations, MetadataMixin):
    """Dataset class for the EmpkinS Dataset.

    This class is the ``tpcp`` dataset class for the EmpkinS dataset. It provides access to the Biopac data (for ECG
    and ICG), the timelogs for the different experimental phases, the reference annotations for the ECG and ICG
    annotations, as well as metadata like age, gender, and BMI.

    Attributes
    ----------
    %(base_attributes_pep)s
    %(base_attributes_pep_label)s
    %(base_attributes_metadata)s
    timelog : :class:`~pandas.DataFrame`
        Timelog data, indicating the start and end of each experimental phase, as a pandas DataFrame.
    labeling_borders : :class:`~pandas.DataFrame`
        Labeling borders for the selected recording as a pandas DataFrame.

    """

    base_path: Path
    use_cache: bool
    exclude_missing_data: bool

    SAMPLING_RATES: ClassVar[dict[str, int]] = {"ecg": 1000, "icg": 1000}

    PHASES: ClassVar[Sequence[str]] = ["Prep", "Pause_1", "Talk", "Math", "Pause_5"]

    CONDITIONS: ClassVar[Sequence[str]] = ["tsst", "ftsst"]

    GENDER_MAPPING: ClassVar[dict[int, str]] = {1: "Female", 2: "Male"}

    def __init__(
        self,
        base_path: path_t,
        groupby_cols: Sequence[str] | None = None,
        subset_index: Sequence[str] | None = None,
        *,
        return_clean: bool = True,
        exclude_missing_data: bool = False,
        use_cache: bool = True,
        only_labeled: bool = False,
        label_type: str = "rater_01",
    ) -> None:
        """Initialize a new ``EmpkinsDataset`` instance.

        Parameters
        ----------
        base_path : :class:`~pathlib.Path` or str
            Path to the root directory of the EmpkinS dataset.
        return_clean : bool
            Whether to return the preprocessed/cleaned ECG and ICG data when accessing the respective properties.
            Default: ``True``.
        exclude_missing_data : bool
            Whether to exclude participants where parts of the data are missing. Default: ``False``.
        use_cache : bool
            Whether to use caching for loading biopac data. Default: ``True``.
        only_labeled : bool
            Whether to only return sections of the biopac data that are labeled (i.e., cut to labeling borders).
            This is necessary when using the dataset for evaluating the performance of PEP extraction algorithms or for
            training ML-based PEP extraction algorithms. Default: ``False``.
        label_type: str, optional
            Which annotations to use. Can be either "rater_01", "rater_02", or "average". Default: "rater_01".

        """
        # ensure pathlib
        self.base_path = base_path
        self.exclude_missing_data = exclude_missing_data
        self.use_cache = use_cache
        self.label_type = label_type
        super().__init__(
            groupby_cols=groupby_cols, subset_index=subset_index, return_clean=return_clean, only_labeled=only_labeled
        )

    def _sanitize_params(self) -> None:
        # ensure pathlib
        self.base_path = Path(self.base_path)

    def create_index(self) -> pd.DataFrame:
        self._sanitize_params()
        # data is located in a folder named "Data" and data per participant is located in folders named "VP_xx"
        participant_ids = [
            participant_dir.name
            for participant_dir in get_subject_dirs(self.base_path.joinpath("data_per_subject"), "VP_[0-9]{3}")
        ]

        # excludes participants where parts of data are missing
        if self.exclude_missing_data:
            for p_id in self.MISSING_DATA:
                if p_id in participant_ids:
                    participant_ids.remove(p_id)

        index = list(product(participant_ids, self.CONDITIONS, self.PHASES))
        index = pd.DataFrame(index, columns=["participant", "condition", "phase"])
        return index

    @property
    def sampling_rate(self) -> dict[str, float]:
        """Return the sampling rates of the ECG and ICG signals.

        Returns
        -------
        dict
            Dictionary with the sampling rates of the ECG and ICG signals in Hz.

        """
        return self.SAMPLING_RATES

    @property
    def sampling_rate_ecg(self) -> int:
        """Return the sampling rate of the ECG signal.

        Returns
        -------
        int
            Sampling rate of the ECG data in Hz.

        """
        return self.SAMPLING_RATES["ecg"]

    @property
    def sampling_rate_icg(self) -> int:
        """Return the sampling rate of the ICG signal.

        Returns
        -------
        int
            Sampling rate of the ICG data in Hz.

        """
        return self.SAMPLING_RATES["icg"]

    @cached_property
    def biopac(self) -> pd.DataFrame:
        """Return the biopac data.

        This method returns the biopac data for the selected participant, condition, and phase. If only one participant
        and condition is selected, the entire dataset is returned. If only one participant, condition, and phase is
        selected, only the data for this phase is returned. In all other cases (i.e., data from multiple participants
        or from multiple conditions), a ``ValueError`` is raised.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Biopac data as a pandas DataFrame.

        Raises
        ------
        ValueError
            If the current subset is not a single participant and condition or
            a single participant, condition, and phase.

        """
        participant = self.index["participant"][0]
        condition = self.index["condition"][0]

        if self.is_single(None):
            phase = self.index["phase"][0]
        elif self.is_single(["participant", "condition"]):
            phase = "all"
        else:
            raise ValueError("Biopac data can only be accessed for one single participant and condition at once!")

        data, fs = self._get_biopac_data(participant, condition, phase)

        if self.only_labeled:
            biopac_data_dict = {}
            labeling_borders = self.labeling_borders

            if self.is_single(None):
                biopac_data_dict = self._cut_to_labeling_borders(data, labeling_borders)
            else:
                for phase in self.PHASES:
                    borders = labeling_borders[labeling_borders["description"].apply(lambda x, ph=phase: ph in x)]
                    biopac_data_dict[phase] = self._cut_to_labeling_borders(data, borders)
            return biopac_data_dict

        return data

    @staticmethod
    def _cut_to_labeling_borders(data: pd.DataFrame, labeling_borders: pd.DataFrame) -> pd.DataFrame:
        start_index = labeling_borders["sample_relative"].iloc[0]
        end_index = labeling_borders["sample_relative"].iloc[-1]
        return data.iloc[start_index:end_index]

    @property
    def icg(self) -> IcgRawDataFrame:
        """Return the ICG channel from the biopac data.

        If ``return_clean`` is set to ``True`` in the ``__init__``, the ICG signal is preprocessed and cleaned using
        the :class:`~biopsykit.signals.icg.preprocessing.IcgPreprocessingBandpass` algorithm before returning it.


        Returns
        -------
        :class:`~pandas.DataFrame`
            ICG data as a pandas DataFrame.

        Raises
        ------
        ValueError
            If the current subset is not a single participant, condition, and phase or a single participant and
            condition.

        """
        if not self.is_single(None):
            raise ValueError(
                "ICG data can only be accessed for a single participant, a single condition, and a single phase!"
            )
        icg = self.biopac[["icg_der"]]
        if self.return_clean:
            algo = IcgPreprocessingBandpass()
            algo.clean(icg=icg, sampling_rate_hz=self.sampling_rate_icg)
            return algo.icg_clean_
        return icg

    @property
    def ecg(self) -> EcgRawDataFrame:
        """Return the ECG channel from the biopac data.

        If ``return_clean`` is set to ``True`` in the ``__init__``, the ECG signal is preprocessed and cleaned using the
        :class:`~biopsykit.signals.ecg.preprocessing.EcgPreprocessingNeurokit` algorithm before returning it.

        Returns
        -------
        :class:`~biopsykit.utils.dtypes.EcgRawDataFrame`
            ECG data as a pandas DataFrame.

        """
        if not self.is_single(None):
            raise ValueError(
                "ECG data can only be accessed for a single participant, a single condition, and a single phase!"
            )
        ecg = self.biopac[["ecg"]]
        if self.return_clean:
            algo = EcgPreprocessingNeurokit()
            algo.clean(ecg=ecg, sampling_rate_hz=self.sampling_rate_ecg)
            return algo.ecg_clean_
        return ecg

    @property
    def timelog(self) -> pd.DataFrame:
        """Return the timelog data.

        The timelog data contains information about the start and end of each experimental phase.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Timelog data as a pandas DataFrame.

        """
        if self.is_single(None):
            participant = self.index["participant"][0]
            condition = self.index["condition"][0]
            phase = self.index["phase"][0]
            return self._get_timelog(participant, condition, phase)

        if self.is_single(["participant", "condition"]):
            if not self._all_phases_selected():
                raise ValueError("Timelog can only be accessed for all phases or one specific phase!")

            participant = self.index["participant"][0]
            condition = self.index["condition"][0]
            return self._get_timelog(participant, condition, "all")

        raise ValueError("Timelog can only be accessed for a single participant and a single condition at once!")

    def _get_biopac_data(self, participant_id: str, condition: str, phase: str) -> tuple[pd.DataFrame, int]:
        if self.use_cache:
            data, fs = _cached_get_biopac_data(self.base_path, participant_id, condition)
        else:
            data, fs = _load_biopac_data(self.base_path, participant_id, condition)

        if phase == "all":
            return data, fs
        # cut biopac data to specified phase
        timelog = self.timelog
        phase_start = timelog[phase]["start"].iloc[0]
        phase_end = timelog[phase]["end"].iloc[0]
        data = data.loc[phase_start:phase_end]
        return data, fs

    def _get_timelog(self, participant_id: str, condition: str, phase: str) -> pd.DataFrame:
        return _load_timelog(self.base_path, participant_id, condition, phase)

    def _all_phases_selected(self) -> bool:
        # check if all phases are selected
        return len(self.index["phase"]) == len(self.PHASES)

    @property
    def labeling_borders(self) -> pd.DataFrame:
        """Return the labeling borders for the selected participant, condition, and phase(s).

        Returns
        -------
        :class:`~pandas.DataFrame`
            Labeling borders as a pandas DataFrame.

        """
        participant = self.index["participant"][0]
        condition = self.index["condition"][0]

        if not self.is_single("participant"):
            raise ValueError("Labeling borders can only be accessed for a single participant.")

        file_path = self.base_path.joinpath(
            f"data_per_subject/{participant}/{condition}/biopac/reference_labels/labeling_borders_{participant}_{condition}.csv"
        )
        data = load_labeling_borders(file_path)

        if self.is_single(None):
            phase = self.index["phase"][0]
            data = data[data["description"].apply(lambda x, ph=phase: ph in x)]

        return data

    @property
    def reference_heartbeats(self) -> pd.DataFrame:
        """Return the reference heartbeats.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference heartbeats as a pandas DataFrame

        """
        return self._load_reference_heartbeats()

    @property
    def reference_labels_ecg(self) -> pd.DataFrame | dict[str, pd.DataFrame]:
        """Return the reference labels for the ECG channel.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ECG channel as a pandas DataFrame.

        """
        return self._load_reference_labels("ECG")

    @property
    def reference_labels_icg(self) -> pd.DataFrame | dict[str, pd.DataFrame]:
        """Return the reference labels for the ICG channel.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ICG channel as a pandas DataFrame.

        """
        return self._load_reference_labels("ICG")

    def _load_reference_heartbeats(self) -> pd.DataFrame:
        reference_ecg = self.reference_labels_ecg
        reference_heartbeats = reference_ecg.reindex(["heartbeat"], level="channel")
        reference_heartbeats = compute_reference_heartbeats(reference_heartbeats)
        return reference_heartbeats

    def _load_reference_labels(self, channel: str) -> pd.DataFrame:
        participant = self.index["participant"][0]
        condition = self.index["condition"][0]
        phases = self.index["phase"]

        if not (self.is_single(None) or len(phases) == len(self.PHASES)):
            raise ValueError(
                "Reference data can only be accessed for a single participant and ALL phases or "
                "for a single participant and a SINGLE phase."
            )

        reference_data_dict = {}
        rater_type = self.label_type
        if self.label_type == "average":
            # TODO implement
            raise NotImplementedError("Average reference labels are not implemented yet.")

        for phase in phases:
            file_path = self.base_path.joinpath(
                f"data_per_subject/{participant}/{condition}/biopac/reference_labels/{rater_type}/"
                f"reference_labels_{participant}_{condition}_{phase.lower()}_{channel.lower()}.csv"
            )
            reference_data = pd.read_csv(file_path)
            reference_data = reference_data.set_index(["heartbeat_id", "channel", "label"])

            start_idx = self.get_subset(phase=phase).labeling_borders.iloc[0]
            reference_data = reference_data.assign(
                sample_relative=reference_data["sample_absolute"] - start_idx["sample_absolute"]
            )

            reference_data_dict[phase] = reference_data

        if self.is_single(None):
            return reference_data_dict[phases[0]]
        return pd.concat(reference_data_dict, names=["phase"])

    @property
    def heartbeats(self) -> HeartbeatSegmentationDataFrame:
        """Segment heartbeats from the ECG data and return the heartbeat borders.

        Returns
        -------
        :class:`~biopsykit.utils.dtypes.HeartbeatSegmentationDataFrame`
            Heartbeats as a pandas DataFrame.

        """
        heartbeat_algo = HeartbeatSegmentationNeurokit(variable_length=True)
        heartbeat_algo.extract(ecg=self.ecg, sampling_rate_hz=self.sampling_rate_ecg)
        heartbeats = heartbeat_algo.heartbeat_list_
        return heartbeats

    @property
    def metadata(self) -> pd.DataFrame:
        """Return metadata for the selected participants.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Metadata as a pandas DataFrame.

        """
        data = pd.read_csv(self.base_path.joinpath("metadata/demographics.csv"))
        data = data.set_index("participant")

        return data.loc[self.index["participant"].unique()]

    @property
    def age(self) -> pd.DataFrame:
        """Return the age of the selected participants.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Age as a pandas DataFrame.

        """
        return self.metadata[["Age"]]

    @property
    def gender(self) -> pd.DataFrame:
        """Return the gender of the selected participants.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Gender as a pandas DataFrame, recoded as {1: "Female", 2: "Male"}

        """
        return self.metadata[["Gender"]].replace(self.GENDER_MAPPING)

    @property
    def bmi(self) -> pd.DataFrame:
        """Compute the BMI of the selected participants and return it.

        Returns
        -------
        :class:`~pandas.DataFrame`
            BMI as a pandas DataFrame

        """
        return bmi(self.metadata[["Weight", "Height"]])



from pathlib import Path

import pandas as pd
from biopsykit.io import load_atimelogger_file
from biopsykit.io.biopac import BiopacDataset

from pepbench.utils._types import path_t


def _build_data_path(base_path: path_t, participant_id: str, condition: str) -> Path:
    data_path = base_path.joinpath(f"data_per_subject/{participant_id}/{condition}")
    assert data_path.exists()
    return data_path


def _load_biopac_data(base_path: path_t, participant_id: str, condition: str) -> tuple[pd.DataFrame, int]:
    biopac_dir_path = _build_data_path(base_path, participant_id=participant_id, condition=condition).joinpath(
        "biopac/raw"
    )

    biopac_file_path = biopac_dir_path.joinpath(f"biopac_data_{participant_id}_{condition}.acq")

    biopac_data = BiopacDataset.from_acq_file(biopac_file_path)
    biopac_df = biopac_data.data_as_df(index="local_datetime")
    fs = next(iter(biopac_data._sampling_rate.values()))
    return biopac_df, fs


def _load_timelog(base_path: path_t, participant_id: str, condition: str, phase: str) -> pd.DataFrame:
    timelog_dir_path = _build_data_path(base_path, participant_id=participant_id, condition=condition).joinpath(
        "timelog/processed"
    )
    timelog_file_path = timelog_dir_path.joinpath(f"{participant_id}_{condition}_processed_timelog.csv")
    timelog = load_atimelogger_file(timelog_file_path, timezone="Europe/Berlin")

    if phase == "all":
        timelog_coarse = timelog.drop("Talk_1", axis=1, level=0)
        timelog_coarse = timelog_coarse.drop("Talk_2", axis=1, level=0)
        timelog_coarse = timelog_coarse.drop("Math_1", axis=1, level=0)
        timelog_coarse = timelog_coarse.drop("Math_2", axis=1, level=0)
        return timelog_coarse
    timelog = timelog.iloc[:, timelog.columns.get_level_values(0) == phase]
    return timelog



from collections.abc import Sequence

import numpy as np
import pandas as pd
from tpcp.validate import FloatAggregator, no_agg

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.evaluation._error_metrics import abs_error, abs_rel_error, error
from pepbench.evaluation._scoring_aggregator import PerSampleAggregator
from pepbench.heartbeat_matching import match_heartbeat_lists
from pepbench.pipelines import BasePepExtractionPipeline

__all__ = ["mean_and_std", "score_pep_evaluation"]


def score_pep_evaluation(pipeline: BasePepExtractionPipeline, datapoint: BasePepDatasetWithAnnotations) -> dict:
    """Run a PEP extraction pipeline on a single datapoint and compute evaluation metrics.

    # *first* averaged over single datapoint and *then* aggregated (mean, std) on total dataset

    The following evaluation metrics are *first* averaged over a single datapoint and *then* aggregated (mean, std)
    on the total dataset:

        * ``pep_reference_ms``: The mean reference PEP in milliseconds.
        * ``pep_estimated_ms``: The mean estimated PEP in milliseconds.
        * ``error_ms``: The mean error between reference and estimated PEP in milliseconds.
        * ``absolute_error_ms``: The mean absolute error between reference and estimated PEP in milliseconds.
        * ``absolute_relative_error_percent``: The mean absolute relative error (relative to the reference PEP)
          between reference and estimated PEP in percent.


    The following evaluation metrics are passed on (since they are single values per datapoint) and then aggregated
    by summing them up (using the `sum_aggregator`):

        * ``num_pep_total``: The total number of PEPs in this datapoint.
        * ``num_pep_valid``: The number of valid PEPs in this datapoint.
        * ``num_pep_invalid``: The number of invalid PEPs in this datapoint.


    The following evaluation metrics are not aggregated but passed on as they are:

        * ``pearson_r``: The Pearson correlation coefficient between the reference and estimated PEP values of all
          matched heartbeats for a single datapoint (excluding NaN values).

    The following evaluation metrics are not aggregated but passed as per-sample values:

        * ``pep_estimation_per_sample``: The estimated and reference PEPs per sample.


    The following evaluation metrics are directly aggregated (mean, std) over *all* samples *without* intermediate
    aggregation on single datapoint:

        * ``error_per_sample_ms``: The mean error between reference and estimated PEP per sample in milliseconds.
        * ``absolute_error_per_sample_ms``: The mean absolute error between reference and estimated PEP per sample
          in milliseconds.
        * ``absolute_relative_error_per_sample_percent``: The mean absolute relative error (relative to the reference
          PEP) between reference and estimated PEP per sample in percent.


    Parameters
    ----------
    pipeline : :class:`pepbench.pipelines.BasePepExtractionPipeline`
        A PEP extraction pipeline. The pipeline must be a subclass of
        :class:`pepbench.pipelines.BasePepExtractionPipeline`.
    datapoint : :class:`pepbench.datasets.BasePepDatasetWithAnnotations`
        A single datapoint from the dataset. The datapoint must be a subclass of
        :class:`pepbench.datasets.BasePepDatasetWithAnnotations`.

    Returns
    -------
    dict
        A dictionary containing the evaluation metrics.

    """
    # not necessary anymore
    # pipeline = pipeline.clone()
    pipeline = pipeline.safe_run(datapoint)
    pep_estimated = pipeline.pep_results_
    pep_reference = datapoint.reference_pep

    heartbeat_matching = match_heartbeat_lists(
        heartbeats_reference=datapoint.reference_heartbeats,
        heartbeats_extracted=pipeline.heartbeat_segmentation_results_,
        tolerance_ms=100,
        sampling_rate_hz=datapoint.sampling_rate_ecg,
    )

    # todo: handle false negatives and false positives, i.e. heartbeats that are not matched
    tp_matches = heartbeat_matching.query("match_type == 'tp'")

    pep_extracted_tp = pep_estimated.loc[tp_matches["heartbeat_id"]]
    pep_reference_tp = pep_reference.loc[tp_matches["heartbeat_id_reference"]]

    pep_merged = _merge_extracted_and_reference_pep(pep_extracted_tp, pep_reference_tp, tp_matches)

    reference_pep_nan_idxs = pep_merged.index[pep_merged[("pep_ms", "reference")].isna()]

    # drop reference PEPs that are NaN
    pep_merged_filter = pep_merged.drop(index=reference_pep_nan_idxs)

    # determine the estimated PEPs that are NaN
    estimated_pep_nan_idxs = pep_merged_filter.index[pep_merged_filter[("pep_ms", "estimated")].isna()]
    num_peps = len(pep_merged_filter)
    num_invalid_peps = len(estimated_pep_nan_idxs)
    num_valid_peps = len(pep_merged_filter) - num_invalid_peps

    pep_reference = pep_merged_filter[("pep_ms", "reference")].to_numpy(na_value=np.nan).astype(float)
    pep_estimated = pep_merged_filter[("pep_ms", "estimated")].to_numpy(na_value=np.nan).astype(float)

    nan_mask = ~np.logical_or(np.isnan(pep_reference), np.isnan(pep_estimated))

    # compute cross-correlation coefficient
    corr_coeff = np.corrcoef(np.compress(nan_mask, pep_reference), np.compress(nan_mask, pep_estimated))
    corr_coeff = corr_coeff[0][1]

    pep_error = error(pep_merged_filter[("pep_ms", "reference")], pep_merged_filter[("pep_ms", "estimated")]).to_numpy(
        na_value=np.nan
    )
    pep_abs_error = abs_error(
        pep_merged_filter[("pep_ms", "reference")], pep_merged_filter[("pep_ms", "estimated")]
    ).to_numpy(na_value=np.nan)
    pep_abs_rel_error = 100 * abs_rel_error(
        pep_merged_filter[("pep_ms", "reference")], pep_merged_filter[("pep_ms", "estimated")]
    ).to_numpy(na_value=np.nan)

    per_sample_aggregator = PerSampleAggregator(mean_and_std)
    sum_aggregator = FloatAggregator(np.nansum)

    return {
        # *first* averaged over single datapoint and *then* aggregated (mean, std) on total dataset
        "pep_reference_ms": np.nanmean(pep_reference),
        "pep_estimated_ms": np.nanmean(pep_estimated),
        "error_ms": np.nanmean(pep_error),
        "absolute_error_ms": np.nanmean(pep_abs_error),
        "absolute_relative_error_percent": np.nanmean(pep_abs_rel_error),
        # no aggregation here necessary since
        "pearson_r": corr_coeff,
        # single values per datapoint => aggregated on total dataset by summing it up (thus the sum_aggregator)
        "num_pep_total": sum_aggregator(num_peps),
        "num_pep_valid": sum_aggregator(num_valid_peps),
        "num_pep_invalid": sum_aggregator(num_invalid_peps),
        # no aggregation, keep per-sample values
        "pep_estimation_per_sample": no_agg(pep_merged_filter),
        # direct aggregation (mean, std) over all samples *without* intermediate aggregation on single datapoint
        "error_per_sample_ms": per_sample_aggregator(pep_error),
        "absolute_error_per_sample_ms": per_sample_aggregator(pep_abs_error),
        "absolute_relative_error_per_sample_percent": per_sample_aggregator(pep_abs_rel_error),
    }


def mean_and_std(vals: Sequence[float]) -> dict:
    return {"mean": float(np.nanmean(vals)), "std": float(np.nanstd(vals))}


def _merge_extracted_and_reference_pep(
    extracted: pd.DataFrame, reference: pd.DataFrame, tp_matches: pd.DataFrame | None = None
) -> pd.DataFrame:
    extracted_original = extracted.copy()
    # if heartbeat_id in index, add it as a column to preserve it in the combined DataFrame
    if "heartbeat_id" in extracted.index.names and "heartbeat_id" in reference.index.names:
        extracted.insert(0, "heartbeat_id", extracted.index.get_level_values("heartbeat_id"))
        reference.insert(0, "heartbeat_id", reference.index.get_level_values("heartbeat_id"))

    common_columns = list(set(reference.columns).intersection(extracted.columns))
    if len(common_columns) == 0:
        raise ValueError("No common columns found in `extracted` and `reference`.")

    extracted = extracted[common_columns]
    reference = reference[common_columns]

    if tp_matches is not None:
        extracted.index = tp_matches.index
        reference.index = tp_matches.index

    matches = extracted.merge(reference, left_index=True, right_index=True, suffixes=("_est", "_ref"))

    # construct MultiIndex columns
    matches.columns = pd.MultiIndex.from_product([["estimated", "reference"], common_columns])
    # make 'metrics' level the uppermost level and sort columns accordingly for readability
    matches = matches.swaplevel(axis=1).sort_index(axis=1, level=0)

    # add rr_intervals to the dataframe; needs to be handled separately since it's not part of the reference data
    rr_interval_extracted = extracted_original[["rr_interval_ms"]].reset_index()
    rr_interval_extracted = pd.concat([rr_interval_extracted, rr_interval_extracted], axis=1)
    rr_interval_extracted.columns = pd.MultiIndex.from_tuples(
        [
            ("heartbeat_id", "estimated"),
            ("rr_interval_ms", "estimated"),
            ("heartbeat_id", "reference"),
            ("rr_interval_ms", "reference"),
        ]
    )
    # drop ("heartbeat_id", "reference")
    rr_interval_extracted = rr_interval_extracted.drop(("heartbeat_id", "reference"), axis=1)

    # add heart_rate_bpm to the dataframe; needs to be handled separately since it's not part of the reference data
    heart_rate_extracted = extracted_original[["heart_rate_bpm"]].reset_index()
    heart_rate_extracted = pd.concat([heart_rate_extracted, heart_rate_extracted], axis=1)
    heart_rate_extracted.columns = pd.MultiIndex.from_tuples(
        [
            ("heartbeat_id", "estimated"),
            ("heart_rate_bpm", "estimated"),
            ("heartbeat_id", "reference"),
            ("heart_rate_bpm", "reference"),
        ]
    )
    # drop ("heartbeat_id", "reference")
    heart_rate_extracted = heart_rate_extracted.drop(("heartbeat_id", "reference"), axis=1)

    # set index to join on the MultiIndex
    matches = matches.set_index(("heartbeat_id", "estimated"))
    rr_interval_extracted = rr_interval_extracted.set_index(("heartbeat_id", "estimated"))
    heart_rate_extracted = heart_rate_extracted.set_index(("heartbeat_id", "estimated"))

    # join the rr_intervals to the DataFrame and reset the index
    matches = matches.join(rr_interval_extracted).join(heart_rate_extracted).reset_index()

    matches = matches.reindex(
        [
            "heartbeat_id",
            "heartbeat_start_sample",
            "heartbeat_end_sample",
            "q_peak_sample",
            "b_point_sample",
            "rr_interval_ms",
            "heart_rate_bpm",
            "pep_sample",
            "pep_ms",
            "nan_reason",
        ],
        axis=1,
        level=0,
    )

    return matches.astype(
        {
            ("heartbeat_id", "estimated"): "Int64",
            ("heartbeat_start_sample", "reference"): "Int64",
            ("heartbeat_start_sample", "estimated"): "Int64",
            ("heartbeat_end_sample", "reference"): "Int64",
            ("heartbeat_end_sample", "estimated"): "Int64",
            ("q_peak_sample", "reference"): "Int64",
            ("q_peak_sample", "estimated"): "Int64",
            ("b_point_sample", "reference"): "Int64",
            ("b_point_sample", "estimated"): "Int64",
            ("rr_interval_ms", "reference"): "Float64",
            ("rr_interval_ms", "estimated"): "Float64",
            ("heart_rate_bpm", "reference"): "Float64",
            ("heart_rate_bpm", "estimated"): "Float64",
            ("pep_sample", "reference"): "Int64",
            ("pep_sample", "estimated"): "Int64",
            ("pep_ms", "reference"): "Float64",
            ("pep_ms", "estimated"): "Float64",
            ("nan_reason", "reference"): "object",
            ("nan_reason", "estimated"): "object",
        }
    )



"""Some custom helper types to make type hints and type checking easier."""

from collections.abc import Hashable, Sequence
from pathlib import Path
from typing import TypeVar

import numpy as np
import pandas as pd

__all__ = ["arr_t", "check_file_exists", "path_t", "str_t"]

_Hashable = Hashable | str

path_t = TypeVar("path_t", str, Path)  # pylint:disable=invalid-name
str_t = TypeVar("str_t", str, Sequence[str])  # pylint:disable=invalid-name
arr_t = TypeVar("arr_t", pd.DataFrame, pd.Series, np.ndarray)  # pylint:disable=invalid-name
T = TypeVar("T")


def check_file_exists(file_path: path_t) -> None:
    # ensure pathlib
    file_path = Path(file_path)
    if not file_path.exists():
        raise FileNotFoundError(f"No file {file_path.absolute()} exists!")



import re
from collections.abc import Sequence
from functools import lru_cache
from itertools import product
from pathlib import Path
from typing import ClassVar

import pandas as pd
from biopsykit.signals.ecg.event_extraction import QPeakExtractionVanLien2013
from biopsykit.signals.ecg.preprocessing import EcgPreprocessingNeurokit
from biopsykit.signals.ecg.segmentation import HeartbeatSegmentationNeurokit
from biopsykit.signals.icg.preprocessing import IcgPreprocessingBandpass
from biopsykit.utils.dtypes import EcgRawDataFrame, IcgRawDataFrame

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.datasets._base_pep_extraction_dataset import base_pep_extraction_docfiller
from pepbench.datasets.time_window_icg._helper import _get_match_heartbeat_label_ids, _load_txt_data
from pepbench.utils._types import path_t

_cached_get_txt_data = lru_cache(maxsize=10)(_load_txt_data)

__all__ = ["TimeWindowIcgDataset"]


@base_pep_extraction_docfiller
class TimeWindowIcgDataset(BasePepDatasetWithAnnotations):
    base_path: Path
    use_cache: bool

    exclude_r_peak_detection_errors: bool

    SAMPLING_RATE: ClassVar[int] = 2000

    PHASES: ClassVar[Sequence[str]] = ["Baseline", "EmotionInduction"]

    SUBSET_R_PEAK_DETECTION_ERRORS: ClassVar[Sequence[str]] = ["IDN_17"]

    def __init__(
        self,
        base_path: path_t,
        groupby_cols: Sequence[str] | None = None,
        subset_index: Sequence[str] | None = None,
        *,
        return_clean: bool = True,
        use_cache: bool = True,
        exclude_r_peak_detection_errors: bool = True,
        only_labeled: bool = False,
    ) -> None:
        self.base_path = base_path
        self.use_cache = use_cache
        self.exclude_r_peak_detection_errors = exclude_r_peak_detection_errors
        self.data_to_exclude = self._find_data_to_exclude()
        super().__init__(
            groupby_cols=groupby_cols, subset_index=subset_index, return_clean=return_clean, only_labeled=only_labeled
        )

    def _sanitize_params(self) -> None:
        # ensure pathlib
        self.base_path = Path(self.base_path)

    def create_index(self) -> pd.DataFrame:
        self._sanitize_params()

        file_list = sorted(self.base_path.joinpath("signals").glob("*.txt"))
        p_ids = [re.findall(r"IDN(\d+).txt", f.name)[0] for f in file_list]
        p_ids = sorted([f"IDN_{p_id.zfill(2)}" for p_id in p_ids])
        index = product(p_ids, self.PHASES)
        index = pd.DataFrame(index, columns=["participant", "phase"])
        index = index.set_index("participant")

        index = index.drop(self.data_to_exclude)
        index = index.reset_index()

        return index

    def _find_data_to_exclude(self) -> Sequence[tuple[str, str]]:
        data_to_exclude = []
        if self.exclude_r_peak_detection_errors:
            data_to_exclude.extend(self.SUBSET_R_PEAK_DETECTION_ERRORS)

        return data_to_exclude

    @property
    def sampling_rate_ecg(self) -> int:
        return self.SAMPLING_RATE

    @property
    def sampling_rate_icg(self) -> int:
        return self.SAMPLING_RATE

    @property
    def data(self) -> pd.DataFrame:
        if not self.is_single("participant"):
            raise ValueError("Data can only be loaded for a single participant.")

        p_id = self.subset_index.iloc[0]["participant"]

        file_path = self.base_path.joinpath("signals").joinpath(f"IDN{int(p_id.split('_')[1])}.txt")

        data = _cached_get_txt_data(file_path) if self.use_cache else _load_txt_data(file_path)

        data = data.copy()
        data.index = pd.to_timedelta(data.index / self.SAMPLING_RATE, unit="s")
        data.index.name = "t"

        if self.is_single("phase"):
            phase = self.subset_index.iloc[0]["phase"]
            split_idx = int(120 * self.sampling_rate_icg)
            # Baseline phase is 0-120s, EmotionInduction phase is 120-end
            data = data.iloc[:split_idx] if phase == "Baseline" else data.iloc[split_idx:]

        return data

    @property
    def ecg(self) -> EcgRawDataFrame:
        """Return the ECG data.

        Returns
        -------
        :class:`~biopsykit.utils.dtypes.EcgRawDataFrame`
            If `return_clean` is `True`, the cleaned ECG data, otherwise the raw ECG data.

        """
        ecg = self.data[["ecg"]]
        if self.return_clean:
            algo = EcgPreprocessingNeurokit()
            algo.clean(ecg=ecg, sampling_rate_hz=self.sampling_rate_ecg)
            ecg = algo.ecg_clean_
        return ecg

    @property
    def icg(self) -> IcgRawDataFrame:
        """Return the ICG data.

        Returns
        -------
        :class:`~biopsykit.utils.dtypes.IcgRawDataFrame`
            If `return_clean` is `True`, the cleaned ICG data, otherwise the raw ICG data.

        """
        icg = self.data[["icg_der"]]
        if self.return_clean:
            algo = IcgPreprocessingBandpass()
            algo.clean(icg=icg, sampling_rate_hz=self.sampling_rate_icg)
            icg = algo.icg_clean_
        return icg

    @property
    def heartbeats(self) -> pd.DataFrame:
        """Segment heartbeats from the ECG data and return the heartbeat borders.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Heartbeats as a pandas DataFrame.

        """
        heartbeat_algo = HeartbeatSegmentationNeurokit()
        heartbeat_algo.extract(ecg=self.ecg, sampling_rate_hz=self.sampling_rate_ecg)
        heartbeats = heartbeat_algo.heartbeat_list_
        return heartbeats

    @property
    def labeling_borders(self) -> pd.DataFrame:
        """Return the labeling borders for a selected participant and phase.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Labeling borders as :class:`~pandas.DataFrame`.

        """
        if not self.is_single("participant"):
            raise ValueError("Labeling borders can only be loaded for a single participant.")

        data = self.data
        labeling_borders = pd.DataFrame({"start_sample": 0, "end_sample": data.shape[0] - 1}, index=[0])
        return labeling_borders

    @property
    def reference_heartbeats(self) -> pd.DataFrame:
        if not self.is_single("participant"):
            raise ValueError("Reference heartbeats can only be loaded for a single participant.")

        reference_heartbeat_folder = self.base_path.joinpath("reference_heartbeats")
        if not reference_heartbeat_folder.exists():
            raise ValueError(
                "Reference heartbeats not found. Please generate them first by calling "
                "`pepbench.datasets.time_window_icg.generate_heartbeat_borders()`."
            )

        p_id = self.subset_index.iloc[0]["participant"]

        file_path = reference_heartbeat_folder.joinpath(f"IDN{int(p_id.split('_')[1])}.csv")
        reference_heartbeats = pd.read_csv(file_path)
        reference_heartbeats = reference_heartbeats.set_index("heartbeat_id")
        if self.is_single("phase"):
            phase = self.subset_index.iloc[0]["phase"]
            split_idx = int(120 * self.sampling_rate_icg)
            if phase == "Baseline":
                # Baseline phase is 0-120s
                reference_heartbeats = reference_heartbeats[reference_heartbeats["start_sample"] <= split_idx]
            else:
                # EmotionInduction phase is 120-end
                reference_heartbeats = reference_heartbeats.loc[reference_heartbeats["start_sample"] > split_idx]
                reference_heartbeats = reference_heartbeats.assign(
                    start_sample=reference_heartbeats["start_sample"] - split_idx,
                    start_time=reference_heartbeats["start_time"] - split_idx / self.sampling_rate_icg,
                    end_sample=reference_heartbeats["end_sample"] - split_idx,
                    r_peak_sample=reference_heartbeats["r_peak_sample"] - split_idx,
                )
                reference_heartbeats.index -= reference_heartbeats.index[0]
        reference_heartbeats = reference_heartbeats.astype(
            {
                "start_sample": "Int64",
                "end_sample": "Int64",
                "r_peak_sample": "Int64",
                "rr_interval_sample": "Int64",
                "rr_interval_ms": "Float64",
            }
        )
        return reference_heartbeats

    @property
    def reference_labels_icg(self) -> pd.DataFrame:
        if not self.is_single("participant"):
            raise ValueError("Reference labels can only be loaded for a single participant.")

        p_id = self.subset_index.iloc[0]["participant"]

        file_path = self.base_path.joinpath("annotations").joinpath(f"IDN{int(p_id.split('_')[1])}.csv")

        reference_heartbeats = self.reference_heartbeats[["start_sample", "end_sample"]]
        b_points = self._load_reference_labels(file_path)
        if self.is_single("phase"):
            phase = self.subset_index.iloc[0]["phase"]
            split_idx = int(120 * self.sampling_rate_icg)
            if phase == "Baseline":
                # Baseline phase is 0-120s
                b_points = b_points[b_points["sample_relative"] <= split_idx]
            else:
                # EmotionInduction phase is 120-end
                b_points = b_points.loc[b_points["sample_relative"] > split_idx]
                b_points -= split_idx
                b_points.index -= b_points.index[0]

        ret = _get_match_heartbeat_label_ids(heartbeats=reference_heartbeats, b_points=b_points)
        # drop duplicate values
        ret = ret.drop_duplicates()

        heartbeats_match = reference_heartbeats.loc[ret]
        heartbeats_match.index.name = "heartbeat_id"
        bpoints_match = b_points.loc[ret.index]
        bpoints_match.index = heartbeats_match.index

        heartbeats_match.columns = ["start", "end"]
        bpoints_match.columns = ["B-point"]

        reference_labels = pd.concat({"heartbeats": heartbeats_match, "ICG": bpoints_match}, axis=1)
        reference_labels = reference_labels.stack([0, 1], future_stack=True)
        reference_labels.index = reference_labels.index.set_names(["heartbeat_id", "channel", "label"])
        # sort values within each heartbeat
        reference_labels = reference_labels.groupby("heartbeat_id", group_keys=False).apply(lambda x: x.sort_values())
        reference_labels = reference_labels.to_frame(name="sample_relative")
        return reference_labels

    @property
    def reference_labels_ecg(self) -> pd.DataFrame:
        """Return the reference labels for the ECG data.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference labels for the ECG data as a pandas DataFrame

        """
        ecg = self.ecg
        q_peak_algo = QPeakExtractionVanLien2013(time_interval_ms=32)
        q_peak_algo.extract(ecg=ecg, heartbeats=self.reference_heartbeats, sampling_rate_hz=self.sampling_rate_ecg)
        q_peaks = q_peak_algo.points_
        q_peaks = q_peaks[["q_peak_sample"]]
        heartbeats = self.reference_heartbeats[["start_sample", "end_sample"]]
        heartbeats.columns = ["start", "end"]
        q_peaks.columns = ["Q-peak"]

        res = pd.concat({"heartbeat": heartbeats, "ECG": q_peaks}, axis=1)
        res = res.stack([0, 1], future_stack=True).sort_values().to_frame(name="sample_relative")
        res.index = res.index.set_names(["heartbeat_id", "channel", "label"])

        return res

    @staticmethod
    def _load_reference_labels(file_path: path_t) -> pd.DataFrame:
        data = pd.read_csv(file_path)
        data = data[["B"]]
        data.columns = ["sample_relative"]
        data.index.name = "heartbeat_id"

        return data



import numpy as np
import pandas as pd

__all__ = ["abs_error", "abs_rel_error", "error", "rel_error"]


def error(ref_data: pd.Series, est_data: pd.Series) -> pd.Series:
    """Calculate the error between the reference and estimated values.

    Parameters
    ----------
    ref_data: :class:`pandas.Series`
        The reference values.
    est_data : :class:`pandas.Series`
        The estimated values.

    Returns
    -------
    error : :class:`pandas.Series`
        The error between the detected and reference values in the form `ref_data` - `est_data`

    """
    return ref_data - est_data


def rel_error(ref_data: pd.Series, est_data: pd.Series) -> pd.Series:
    """Calculate the relative error between the reference and estimated values.

    Parameters
    ----------
    ref_data : :class:`pandas.Series`
        The reference values.
    est_data : :class:`pandas.Series`
        The estimated values.

    Returns
    -------
    error : :class:`pandas.Series`
        The relative error between the reference and estimated values in the form (`ref_data` - `est_data`) / `ref_data`

    """
    result = (ref_data - est_data) / ref_data
    result = result.replace([np.inf, -np.inf], pd.NA)
    return result


def abs_error(ref_data: pd.Series, est_data: pd.Series) -> pd.Series:
    """Calculate the absolute error between the reference and estimated values.

    Parameters
    ----------
    ref_data : :class:`pandas.Series`
        The reference values.
    est_data : :class:`pandas.Series`
        The estimated values.

    Returns
    -------
    error : :class:`pandas.Series`
        The absolute error between the reference and estimated values in the
        form `abs(ref_data - est_data)`

    """
    return np.abs(ref_data - est_data)


def abs_rel_error(ref_data: pd.Series, est_data: pd.Series) -> pd.Series:
    """Calculate the absolute relative error between the reference and estimated values.

    Parameters
    ----------
    ref_data : :class:`pandas.Series`
        The reference values.
    est_data : :class:`pandas.Series`
        The estimated values.

    Returns
    -------
    error : :class:`pandas.Series`
        The absolute relative error between the reference and estimated values in the
        form `abs((ref_data - est_data) / ref_data)`

    """
    result = np.abs((ref_data - est_data) / ref_data)
    result = result.replace([np.inf, -np.inf], pd.NA)
    return result



import json
import warnings
from collections import namedtuple
from collections.abc import Callable
from typing import Any

import numpy as np
import pandas as pd
from tpcp import Algorithm
from tpcp.validate import FloatAggregator, Scorer, validate
from typing_extensions import Self

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.evaluation._scoring import mean_and_std, score_pep_evaluation
from pepbench.pipelines import BasePepExtractionPipeline
from pepbench.utils._timing import measure_time
from pepbench.utils._types import path_t

__all__ = ["ChallengeResults", "PepEvaluationChallenge"]

ChallengeResults = namedtuple("ChallengeResults", ["agg_mean_std", "agg_total", "single", "per_sample"])


# TODO add CrossValidateChallenge
class PepEvaluationChallenge(Algorithm):
    """Evaluation challenge for PEP extraction pipelines.

    This is the ``tpcp`` implementation of the evaluation challenge for PEP extraction pipelines.
    It is used to evaluate the performance of a PEP extraction pipeline on a given dataset.

    """

    _action_methods = "run"

    dataset: BasePepDatasetWithAnnotations
    scoring: Callable | None

    results_: dict
    results_agg_mean_std_: pd.DataFrame
    results_agg_total_: pd.DataFrame
    results_single_: pd.DataFrame
    results_per_sample_: pd.DataFrame

    timing_information_: dict
    # timing information
    start_time_utc_timestamp_: float
    start_time_utc: str
    end_time_utc_timestamp_: float
    end_time_: str
    runtime_s_: float

    def __init__(
        self,
        *,
        dataset: BasePepDatasetWithAnnotations,
        scoring: Callable = score_pep_evaluation,
        validate_kwargs: dict | None = None,
    ) -> None:
        """Initialize a new evaluation challenge.

        To initialize a new evaluation challenge, you need to provide a dataset and a scoring function. Afterwards,
        you can challenge a specific PEP extraction pipeline by passing it to the ``run`` method.

        Parameters
        ----------
        dataset : BasePepDatasetWithAnnotations
            The dataset to evaluate the pipeline on. The dataset needs to be a subclass of
            ``BaseUnifiedPepExtractionDataset``, which provides the necessary unified interface to access the data.
        scoring : Callable, optional
            The scoring function to use for the evaluation. The scoring function should take the pipeline and a
            datapoint from the dataset as input and return a dictionary with the evaluation results. The default
            scoring function is :func:``pepbench.evaluation.score_pep_evaluation``.
        validate_kwargs : dict, optional
            Additional keyword arguments to pass to the :class:``tpcp.validate.Scorer`` class.

        """
        self.dataset = dataset
        self.scoring = scoring
        self.validate_kwargs = validate_kwargs or {}

    def run(self, pipeline: BasePepExtractionPipeline) -> Self:
        """Run the evaluation challenge for a given pipeline.

        Parameters
        ----------
        pipeline : BasePepExtractionPipeline
            The PEP extraction pipeline to evaluate. The pipeline needs to be a subclass of
            :class:``pepbench.pipelines.BasePepExtractionPipeline`` and should be able to process the dataset.

        Returns
        -------
        Self

        """
        with measure_time() as timing_results, warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)  # better specify the RuntimeWarning using regex
            mean_std_agg = FloatAggregator(mean_and_std)
            scorer = Scorer(score_pep_evaluation, default_aggregator=mean_std_agg, **self.validate_kwargs)
            self.results_ = validate(pipeline=pipeline, dataset=self.dataset, scoring=scorer)
            self.results_as_df()

        self.timing_information_ = timing_results
        self._set_attrs_from_dict(timing_results)
        return self

    def save_results(self, folder_path: path_t, filename_stub: str) -> None:
        """Save the results of the evaluation to disk.

        Parameters
        ----------
        folder_path : :class:``pathlib.Path`` or str
            The folder path to save the results to.
        filename_stub : str
            The filename stub to use for the results file.

        """
        timing_information_path = folder_path.joinpath(f"{filename_stub}_timing_information.json")
        # save timing information as json
        with timing_information_path.open("w") as fp:
            json.dump(self.timing_information_, fp)

        # save agg results as csv
        results_agg_mean_std_path = folder_path.joinpath(f"{filename_stub}_results_agg_mean_std.csv")
        self.results_agg_mean_std_.to_csv(results_agg_mean_std_path)

        results_agg_total_path = folder_path.joinpath(f"{filename_stub}_results_agg_total.csv")
        self.results_agg_total_.to_csv(results_agg_total_path)

        # save single results as csv
        results_single_path = folder_path.joinpath(f"{filename_stub}_results_single.csv")
        self.results_single_.to_csv(results_single_path)

        # save per sample results as csv
        results_per_sample_path = folder_path.joinpath(f"{filename_stub}_results_per-sample.csv")
        self.results_per_sample_.to_csv(results_per_sample_path)

    def _set_attrs_from_dict(self, attr_dict: dict[str, Any]) -> None:
        """Set attributes of an object from a dictionary.

        Parameters
        ----------
        obj
            The object to set the attributes on.
        attr_dict
            The dictionary with the attributes to set.
        """
        for key, value in attr_dict.items():
            setattr(self, f"{key}_", value)

    def results_as_df(self) -> Self:
        """Convert the results to pandas DataFrames.

        The results are stored as attributes on the object. The following results are created:
            * ``results_agg_mean_std_``: The mean and standard deviation of the aggregated results.
            * ``results_agg_total_``: The total number of valid and invalid PEPs.
            * ``results_single_``: The single (non-aggregated) results for each datapoint.
            * ``results_per_sample_``: The per-sample results for each datapoint.

        Returns
        -------
        Self

        """
        results = self.results_.copy()

        data_labels = results["data_labels"]
        subset = self.dataset.get_subset(group_labels=data_labels[0])

        results_subset_single = {
            key.replace("single__", ""): val[0]
            for key, val in self.results_.items()
            if key.startswith("single__") and "per_sample" not in key
        }
        result_df_single = pd.DataFrame.from_dict(results_subset_single)
        result_df_single.index = pd.MultiIndex.from_frame(subset.index)

        results_subset_agg = {
            key.replace("agg__", ""): val[0] for key, val in results.items() if key.startswith("agg__")
        }
        # mean and std aggregations
        results_subset_agg_mean_std = {
            agg_type: {
                key.replace(f"__{agg_type}", ""): val
                for key, val in results_subset_agg.items()
                if key.endswith(f"__{agg_type}")
            }
            for agg_type in ["mean", "std"]
        }
        # add "total" aggregation
        results_subset_agg_total = {"total": {key: val for key, val in results_subset_agg.items() if "num_" in key}}

        result_df_agg_mean_std = pd.DataFrame.from_dict(results_subset_agg_mean_std)
        result_df_agg_mean_std.index.name = "metrics"

        result_df_agg_total = pd.DataFrame.from_dict(results_subset_agg_total)
        result_df_agg_total = result_df_agg_total.astype(int)
        result_df_agg_total = result_df_agg_total.reindex(["num_pep_total", "num_pep_valid", "num_pep_invalid"])
        result_df_agg_total.index.name = "metrics"

        results_subset_per_sample = {
            key.replace("single__", ""): val[0]
            for key, val in results.items()
            if key.startswith("single__") and "per_sample" in key
        }
        # concatenate the per_sample results
        pep_estimation = results_subset_per_sample.pop("pep_estimation_per_sample")
        pep_estimation = {
            tuple(key): test_idx for key, test_idx in zip(subset.index.to_numpy(), pep_estimation, strict=False)
        }
        pep_estimation = pd.concat(pep_estimation)
        pep_estimation.index.names = [*list(subset.index.columns), ""]
        results_subset_per_sample = {key: np.concatenate(val, axis=0) for key, val in results_subset_per_sample.items()}

        result_df_per_sample = pd.DataFrame.from_dict(results_subset_per_sample)
        result_df_per_sample.columns = pd.MultiIndex.from_product([list(result_df_per_sample.columns), ["metric"]])
        result_df_per_sample.index = pep_estimation.index
        result_df_per_sample = pd.concat([pep_estimation, result_df_per_sample], axis=1)

        self._set_attrs_from_dict(
            {
                "results_agg_mean_std": result_df_agg_mean_std,
                "results_agg_total": result_df_agg_total,
                "results_single": result_df_single,
                "results_per_sample": result_df_per_sample,
            }
        )
        return self



"""Module for utility functions and classes."""

from pepbench.utils import exceptions, styling
from pepbench.utils._rename_maps import get_nan_reason_mapping, rename_algorithms, rename_metrics

__all__ = ["exceptions", "get_nan_reason_mapping", "rename_algorithms", "rename_metrics", "styling"]



"""Module for ICG event extraction algorithms."""

from biopsykit.signals.icg.event_extraction import (
    BPointExtractionArbol2017IsoelectricCrossings,
    BPointExtractionArbol2017SecondDerivative,
    BPointExtractionArbol2017ThirdDerivative,
    BPointExtractionDebski1993SecondDerivative,
    BPointExtractionDrost2022,
    BPointExtractionForouzanfar2018,
    BPointExtractionLozano2007LinearRegression,
    BPointExtractionLozano2007QuadraticRegression,
    BPointExtractionMiljkovic2022,
    BPointExtractionPale2021,
    BPointExtractionSherwood1990,
    BPointExtractionStern1985,
    CPointExtractionScipyFindPeaks,
)

__all__ = [
    "BPointExtractionArbol2017IsoelectricCrossings",
    "BPointExtractionArbol2017SecondDerivative",
    "BPointExtractionArbol2017ThirdDerivative",
    "BPointExtractionDebski1993SecondDerivative",
    "BPointExtractionDrost2022",
    "BPointExtractionForouzanfar2018",
    "BPointExtractionLozano2007LinearRegression",
    "BPointExtractionLozano2007QuadraticRegression",
    "BPointExtractionMiljkovic2022",
    "BPointExtractionPale2021",
    "BPointExtractionSherwood1990",
    "BPointExtractionStern1985",
    "CPointExtractionScipyFindPeaks",
]



"""Tools for handling and analyzing annotation data.

This package includes functions for:
- Computing annotation differences,
- Loading and matching annotations from datasets,
- Normalizing annotations to heartbeat start times,
- Statistical analysis of annotation differences, including binning and computing descriptive statistics,
- Calculating the Intraclass Correlation Coefficient (ICC), and
- Integrating annotation agreement information into results dataframes.

"""

from pepbench.annotations._annotations import (
    compute_annotation_differences,
    load_annotations_from_dataset,
    match_annotations,
    normalize_annotations_to_heartbeat_start,
)

__all__ = [
    "compute_annotation_differences",
    "load_annotations_from_dataset",
    "match_annotations",
    "normalize_annotations_to_heartbeat_start",
]



"""Functions for analyzing and processing annotation differences.

The functions include utilities for computing descriptive statistics, binning annotation differences, calculating the
Intraclass Correlation Coefficient (ICC), and integrating annotation agreement information into results dataframes.

"""

__all__ = [
    "add_annotation_agreement_to_results_dataframe",
    "bin_annotation_differences",
    "compute_icc",
    "describe_annotation_differences",
]

from collections.abc import Sequence

import pandas as pd
import pingouin as pg

from pepbench.annotations._annotations import compute_annotation_differences, normalize_annotations_to_heartbeat_start
from pepbench.data_handling import add_unique_id_to_results_dataframe


def describe_annotation_differences(annotation_diffs: pd.DataFrame, include_absolute: bool = True) -> pd.DataFrame:
    """Generate descriptive statistics for annotation differences.

    This function computes descriptive statistics for the provided annotation differences dataframe.
    Optionally, it can include the absolute values of the differences as an additional column.

    Parameters
    ----------
    annotation_diffs : :class:`~pandas.DataFrame`
        A dataframe containing annotation differences with a column named "difference_ms".
    include_absolute : bool, optional
        If True, includes the absolute values of the differences in the descriptive statistics.
        Default is True.

    Returns
    -------
    :class:`~pandas.DataFrame`
        A transposed dataframe containing descriptive statistics for the annotation differences.
    """
    annotation_diffs_describe = annotation_diffs.copy()
    if include_absolute:
        annotation_diffs_describe = annotation_diffs_describe.assign(
            difference_ms_absolute=annotation_diffs_describe["difference_ms"].abs()
        )
    return annotation_diffs_describe.describe().T


def bin_annotation_differences(
    annotation_diffs: pd.DataFrame, bins: Sequence[int] | None = None, labels: Sequence[str] | None = None
) -> pd.DataFrame:
    """Bin annotation differences into specified categories.

    This function categorizes annotation differences into bins and assigns labels to each bin.
    If no bins are provided, default bins are used. The resulting bins are returned as a dataframe.

    Parameters
    ----------
    annotation_diffs : :class:`~pandas.DataFrame`
        A dataframe containing annotation differences to be binned.
    bins : list of int, optional
        A sequence of bin edges. If not provided, default bins [0, 4, 10] are used.
    labels : list of str, optional
        A sequence of labels corresponding to the bins. If not provided, no labels are assigned.

    Returns
    -------
    :class:`~pandas.DataFrame`
        A dataframe with a single column named "annotation_bins" containing the binned annotation differences.
    """
    if bins is None:
        bins = [0, 4, 10]
    annotation_bins = pd.cut(
        annotation_diffs.abs().squeeze(),
        bins=[*bins, annotation_diffs.max().squeeze()],
        include_lowest=True,
        labels=labels,
    )
    return annotation_bins.to_frame(name="annotation_bins")


def compute_icc(annotation_diffs: pd.DataFrame, sampling_rate_hz: float) -> pd.DataFrame:
    """Compute the Intraclass Correlation Coefficient (ICC) for annotation differences.

    This function normalizes annotation differences to the heartbeat start, adds unique IDs
    for each rater, and calculates the ICC using the Pingouin library.

    Parameters
    ----------
    annotation_diffs : :class:`~pandas.DataFrame`
        A dataframe containing annotation differences.
    sampling_rate_hz : float
        The sampling rate in Hertz used to normalize annotation differences.

    Returns
    -------
    :class:`~pandas.DataFrame`
        A dataframe containing the computed ICC values.
    """
    annotation_diffs_normalized = normalize_annotations_to_heartbeat_start(
        annotation_diffs, sampling_rate_hz=sampling_rate_hz
    )
    annotation_diffs_normalized = add_unique_id_to_results_dataframe(annotation_diffs_normalized, algo_levels=["rater"])

    return pg.intraclass_corr(
        data=annotation_diffs_normalized.reset_index(),
        targets="id_concat",
        ratings="difference_ms",
        raters="rater",
        nan_policy="omit",
    )


def add_annotation_agreement_to_results_dataframe(
    results_per_sample: pd.DataFrame, annotations: pd.DataFrame, sampling_rate_hz: float
) -> pd.DataFrame:
    """Add annotation agreement information to the results dataframe.

    This function computes annotation differences, bins them into categories (e.g., high, medium, low),
    and integrates this information into the provided results dataframe. The resulting dataframe
    includes an additional index level for annotation agreement bins.

    Parameters
    ----------
    results_per_sample : :class:`~pandas.DataFrame`
        A dataframe containing results for each sample.
    annotations : :class:`~pandas.DataFrame`
        A dataframe containing annotation data to compute differences.
    sampling_rate_hz : float
        The sampling rate in Hertz used to normalize annotation differences.

    Returns
    -------
    :class:`~pandas.DataFrame`
        A dataframe with the original results and added annotation agreement information.
    """
    annotation_diffs = compute_annotation_differences(annotations, sampling_rate_hz=sampling_rate_hz)
    annotation_bins = bin_annotation_differences(annotation_diffs, labels=["high", "medium", "low"])

    annotation_bins.index = annotation_bins.index.rename({"heartbeat_id": "id"})
    annotation_bins = pd.concat(
        {"Annotation Agreement": pd.concat({"annotation_agreement": annotation_bins}, axis=1)}, axis=1
    )

    results_per_sample = results_per_sample.join(annotation_bins)

    results_per_sample = results_per_sample.reindex(
        ["absolute_error_per_sample_ms", "annotation_agreement"], level=1, axis=1
    ).set_index(("Annotation Agreement", "annotation_agreement", "annotation_bins"), append=True)
    results_per_sample.index = results_per_sample.index.rename(
        {("Annotation Agreement", "annotation_agreement", "annotation_bins"): "agreement_bins"}
    )
    return results_per_sample



import pandas as pd
from tqdm.auto import tqdm

from pepbench.datasets import BasePepDatasetWithAnnotations
from pepbench.heartbeat_matching import match_heartbeat_lists

__all__ = [
    "compute_annotation_differences",
    "load_annotations_from_dataset",
    "match_annotations",
    "normalize_annotations_to_heartbeat_start",
]


def load_annotations_from_dataset(
    dataset_01: BasePepDatasetWithAnnotations, dataset_02: BasePepDatasetWithAnnotations
) -> pd.DataFrame:
    labels_ecg_dict = {}
    labels_icg_dict = {}
    for subset_01, subset_02 in tqdm(list(zip(dataset_01.groupby(None), dataset_02.groupby(None), strict=False))):
        labels_ecg = match_annotations(
            subset_01.reference_labels_ecg, subset_02.reference_labels_ecg, dataset_01.sampling_rate_ecg
        )
        labels_icg = match_annotations(
            subset_01.reference_labels_icg, subset_02.reference_labels_icg, dataset_02.sampling_rate_ecg
        )

        labels_ecg_dict[subset_01.group_label] = labels_ecg
        labels_icg_dict[subset_01.group_label] = labels_icg

    labels_ecg_total = pd.concat(labels_ecg_dict, names=dataset_01.group_labels[0]._fields)
    labels_icg_total = pd.concat(labels_icg_dict, names=dataset_01.group_labels[0]._fields)

    labels_ecg_total = labels_ecg_total.xs("sample_relative", level="sample", axis=1)
    labels_icg_total = labels_icg_total.xs("sample_relative", level="sample", axis=1)

    return pd.concat({"ECG": labels_ecg_total, "ICG": labels_icg_total}, names=["signal"])


def match_annotations(
    annotations_01: pd.DataFrame, annotations_02: pd.DataFrame, sampling_rate_hz: float
) -> pd.DataFrame:
    heartbeats_01 = annotations_01.unstack("label")["sample_relative"][["start", "end"]].dropna()
    # heartbeats_01 = annotations_01.reindex(["start", "end"], level="label")["sample_relative"].unstack()
    heartbeats_01 = heartbeats_01.droplevel(-1)
    heartbeats_01.columns = ["start_sample", "end_sample"]

    heartbeats_02 = annotations_02.unstack("label")["sample_relative"][["start", "end"]].dropna()
    # heartbeats_02 = annotations_02.reindex(["start", "end"], level="label")["sample_relative"].unstack()
    heartbeats_02 = heartbeats_02.droplevel(-1)
    heartbeats_02.columns = ["start_sample", "end_sample"]

    heartbeat_matching = match_heartbeat_lists(
        heartbeats_reference=heartbeats_01,
        heartbeats_extracted=heartbeats_02,
        tolerance_ms=100,
        sampling_rate_hz=sampling_rate_hz,
    )

    tp_matches = heartbeat_matching.query("match_type == 'tp'")

    annotations_01_tp = annotations_01.loc[tp_matches["heartbeat_id_reference"]]
    annotations_02_tp = annotations_02.loc[tp_matches["heartbeat_id"]]
    annotations_01_tp = annotations_01_tp.reset_index(["channel", "label"])
    annotations_02_tp = annotations_02_tp.reset_index(["channel", "label"])

    annotations_02_tp.index = annotations_01_tp.index

    annotations_01_tp_new = annotations_01_tp.set_index(["channel", "label"], append=True)
    annotations_02_tp_new = annotations_02_tp.set_index(["channel", "label"], append=True)
    annotations_01_tp_new = annotations_01_tp_new.dropna()
    annotations_02_tp_new = annotations_02_tp_new.dropna()

    annotations = pd.concat(
        {"rater_01": annotations_01_tp_new, "rater_02": annotations_02_tp_new}, names=["rater", "sample"], axis=1
    )

    return annotations


def compute_annotation_differences(annotations: pd.DataFrame, sampling_rate_hz: float | None = None) -> pd.DataFrame:
    """Compute the difference in samples between the two raters."""
    if annotations.columns.nlevels == 1:
        annotations = annotations["rater_01"] - annotations["rater_02"]
    else:
        annotations = annotations["rater_01"]["sample_relative"] - annotations["rater_02"]["sample_relative"]

    if sampling_rate_hz is not None:
        annotations = annotations / sampling_rate_hz * 1000
        annotations = annotations.to_frame("difference_ms")
    else:
        annotations = annotations.to_frame("difference_samples")
    annotations = annotations.drop("heartbeat", level="channel")
    annotations = annotations.drop("Artefact", level="label")
    annotations = annotations.droplevel(["label", "channel"])

    return annotations


def normalize_annotations_to_heartbeat_start(
    annotations: pd.DataFrame, sampling_rate_hz: float | None = None
) -> pd.DataFrame:
    """Normalize the annotations to the start of the heartbeat."""
    annotations = annotations.drop("end", level="label").drop("Artefact", level="label")
    annotations = annotations.droplevel("channel").unstack()
    annotations = annotations.T.groupby("rater").diff().dropna(how="all")
    annotations = annotations.droplevel("label")
    annotations = annotations.T
    annotations = annotations.stack()
    if sampling_rate_hz is not None:
        annotations = annotations / sampling_rate_hz * 1000
        annotations = annotations.to_frame("difference_ms")
    else:
        annotations = annotations.to_frame("difference_samples")

    return annotations



import pandas as pd
from biopsykit.utils.dtypes import EcgRawDataFrame, HeartbeatSegmentationDataFrame, IcgRawDataFrame
from tpcp import Dataset

from pepbench._docutils import make_filldoc

base_pep_extraction_docfiller = make_filldoc(
    {
        "base_attributes_pep": """
        icg : :class:`~biopsykit.utils.dtypes.IcgRawDataFrame`
            The raw ICG data.
        ecg : :class:`~biopsykit.utils.dtypes.EcgRawDataFrame`
            The raw ECG data.
        sampling_rate_ecg : int
            The sampling rate of the ECG data in Hz.
        sampling_rate_icg : int
            The sampling rate of the ICG data in Hz.
        heartbeats : :class:`~biopsykit.utils.dtypes.HeartbeatSegmentationDataFrame`
            The heartbeats extracted from the ECG data.
        """,
        "base_attributes_metadata": """
        age : :class:`~pandas.DataFrame`
            The age of the participants
        gender : :class:`~pandas.DataFrame`
            The gender of the participants
        bmi : :class:`~pandas.DataFrame`
            The BMI of the participants
        metadata : :class:`~pandas.DataFrame`
            The metadata of the participants, consisting of a combination of age, gender, and BMI.
        """,
        "base_attributes_pep_label": """
        reference_pep : :class:`~pandas.DataFrame`
            The reference PEP data.
        reference_heartbeats : :class:`~pandas.DataFrame`
            The reference heartbeats.
        reference_labels_ecg : :class:`~pandas.DataFrame`
            The reference labels for the ECG data.
        reference_labels_icg : :class:`~pandas.DataFrame`
            The reference labels for the ICG data.
        """,
    }
)


@base_pep_extraction_docfiller
class BasePepDataset(Dataset):
    """Interface for all datasets for PEP extraction from ICG and ECG data.

    This class defines the interface for datasets that are used for PEP extraction using the
    :class:`~pepbench.pipelines.PepExtractionPipeline`. It provides the necessary properties and methods to access the
    data and metadata required for PEP extraction. It is not intended to be used directly, but rather as a base class
    for other datasets.

    Attributes
    ----------
    %(base_attributes_pep)s

    """

    return_clean: bool

    def __init__(
        self,
        groupby_cols: list[str] | str | None = None,
        subset_index: pd.DataFrame | None = None,
        return_clean: bool = True,
    ) -> None:
        self.return_clean = return_clean
        super().__init__(groupby_cols=groupby_cols, subset_index=subset_index)

    @property
    def icg(self) -> IcgRawDataFrame:
        """The raw ICG data.

        Returns
        -------
        :class:`~biopsykit.utils.dtypes.IcgRawDataFrame`
            The raw ICG data.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def ecg(self) -> EcgRawDataFrame:
        """The raw ECG data.

        Returns
        -------
        :class:`~biopsykit.utils.dtypes.EcgRawDataFrame`
            The raw ECG data.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def sampling_rate_ecg(self) -> int:
        """The sampling rate of the ECG data in Hz.

        Returns
        -------
        int
            The sampling rate of the ECG data in Hz.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def sampling_rate_icg(self) -> int:
        """The sampling rate of the ICG data in Hz.

        Returns
        -------
        int
            The sampling rate of the ICG data in Hz.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def heartbeats(self) -> HeartbeatSegmentationDataFrame:
        """The heartbeats extracted from the ECG data.

        Returns
        -------
        :class:`~biopsykit.utils.dtypes.HeartbeatSegmentationDataFrame`
            The heartbeats extracted from the ECG data.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")


class MetadataMixin(Dataset):
    """Interface for all datasets that contain certain metadata.

    This interface can be used by datasets that contain metadata like age, gender, and BMI.

    Attributes
    ----------
    base_demographics : :class:`~pandas.DataFrame`
        The base demographics of the participants, including gender, age, and BMI.
    %(base_attributes_metadata)s

    """

    def __init__(self, groupby_cols: list[str] | str | None = None, subset_index: pd.DataFrame | None = None) -> None:
        super().__init__(groupby_cols=groupby_cols, subset_index=subset_index)

    @property
    def base_demographics(self) -> pd.DataFrame:
        return pd.concat([self.gender, self.age, self.bmi], axis=1)

    @property
    def age(self) -> pd.DataFrame:
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def gender(self) -> pd.DataFrame:
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def bmi(self) -> pd.DataFrame:
        raise NotImplementedError("This property needs to be implemented in the subclass!")


class PepLabelMixin(Dataset):
    """Interface for all datasets with manually labeled PEP data.

    This interface can be used by datasets that contain manually labeled PEP data. It provides the necessary properties
    to access the reference PEP data and the reference heartbeats.

    Attributes
    ----------
    %(base_attributes_pep_label)s

    """

    def __init__(self, groupby_cols: list[str] | str | None = None, subset_index: pd.DataFrame | None = None) -> None:
        super().__init__(groupby_cols=groupby_cols, subset_index=subset_index)

    @property
    def reference_heartbeats(self) -> pd.DataFrame:
        """The reference heartbeats.

        Returns
        -------
        :class:`~pandas.DataFrame`
            The reference heartbeats.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def reference_labels_ecg(self) -> pd.DataFrame | dict[str, pd.DataFrame]:
        """The reference labels for the ECG data.

        Returns
        -------
        :class:`~pandas.DataFrame`
            The reference labels for the ECG data.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")

    @property
    def reference_labels_icg(self) -> pd.DataFrame | dict[str, pd.DataFrame]:
        """The reference labels for the ICG data.

        Returns
        -------
        :class:`~pandas.DataFrame`
            The reference labels for the ICG data.

        """
        raise NotImplementedError("This property needs to be implemented in the subclass!")


@base_pep_extraction_docfiller
class BasePepDatasetWithAnnotations(BasePepDataset, PepLabelMixin):
    """Unified interface for datasets used for evaluating PEP extraction algorithms.

    This interface extends the :class:`~pepbench.datasets.BasePepDataset` by adding support for metadata and
    reference PEP data.

    Attributes
    ----------
    %(base_attributes_pep)s
    %(base_attributes_pep_label)s
    %(base_attributes_metadata)s

    """

    only_labeled: bool

    def __init__(
        self,
        groupby_cols: list[str] | str | None = None,
        subset_index: pd.DataFrame | None = None,
        return_clean: bool = True,
        only_labeled: bool = False,
    ) -> None:
        super().__init__(groupby_cols=groupby_cols, subset_index=subset_index, return_clean=return_clean)
        self.only_labeled = only_labeled

    @property
    def reference_pep(self) -> pd.DataFrame:
        """Return the reference PEP values.

        Returns
        -------
        :class:`~pandas.DataFrame`
            Reference PEP values as a pandas DataFrame

        """
        """Compute the reference PEP values between the reference Q-peak and B-point labels.

            Parameters
            ----------
            subset : :class:`pepbench.datasets.BasePepDatasetWithAnnotations`
                Subset of a dataset containing the reference labels.

            Returns
            -------
            :class:`pandas.DataFrame`
                DataFrame containing the computed PEP values.

            """
        heartbeats = self.reference_heartbeats
        reference_icg = self.reference_labels_icg
        reference_ecg = self.reference_labels_ecg

        b_points = reference_icg.reindex(["ICG", "Artefact"], level="channel").droplevel("label")
        b_points = self._fill_unlabeled_artefacts(b_points, reference_icg)
        b_point_artefacts = b_points.reindex(["Artefact"], level="channel").droplevel("channel")
        b_points = b_points.reindex(["ICG"], level="channel").droplevel("channel")

        q_peaks = reference_ecg.reindex(["ECG", "Artefact"], level="channel").droplevel("label")
        q_peaks = self._fill_unlabeled_artefacts(q_peaks, reference_ecg)
        q_peak_artefacts = q_peaks.reindex(["Artefact"], level="channel").droplevel("channel")
        q_peaks = q_peaks.reindex(["ECG"], level="channel").droplevel("channel")

        pep_reference = heartbeats.copy()
        pep_reference.columns = [
            f"heartbeat_{col}" if col != "r_peak_sample" else "r_peak_sample" for col in heartbeats.columns
        ]

        pep_reference = pep_reference.assign(
            q_peak_sample=q_peaks["sample_relative"],
            b_point_sample=b_points["sample_relative"],
            nan_reason=pd.NA,
        )
        # fill nan_reason column with artefact information
        pep_reference.loc[b_point_artefacts.index, "nan_reason"] = "icg_artefact"
        pep_reference.loc[q_peak_artefacts.index, "nan_reason"] = "ecg_artefact"

        pep_reference = pep_reference.assign(
            pep_sample=pep_reference["b_point_sample"] - pep_reference["q_peak_sample"]
        )
        pep_reference = pep_reference.assign(pep_ms=pep_reference["pep_sample"] / self.sampling_rate_ecg * 1000)

        # reorder columns
        pep_reference = pep_reference[
            [
                "heartbeat_start_sample",
                "heartbeat_end_sample",
                "q_peak_sample",
                "b_point_sample",
                "pep_sample",
                "pep_ms",
                "nan_reason",
            ]
        ]

        return pep_reference.convert_dtypes(infer_objects=True)

    @staticmethod
    def _fill_unlabeled_artefacts(points: pd.DataFrame, reference_data: pd.DataFrame) -> pd.DataFrame:
        # get the indices of reference_icg that are not in b_points.index => they are artefacts but were not labeled
        heartbeat_ids = reference_data.index.get_level_values("heartbeat_id").unique()
        # insert "Artefact" label for artefacts that were not labeled to b_points,
        # set the sample to the middle of the heartbeat
        artefact_ids = list(heartbeat_ids.difference(points.droplevel("channel").index))
        for artefact_id in artefact_ids:
            start_abs, end_abs = reference_data.xs(artefact_id, level="heartbeat_id")["sample_absolute"]
            start_rel, end_rel = reference_data.xs(artefact_id, level="heartbeat_id")["sample_relative"]
            points.loc[(artefact_id, "Artefact"), :] = (int((start_abs + end_abs) / 2), int((start_rel + end_rel) / 2))

        points = points.sort_index()
        return points
